import typing as t
import subprocess
import logging
import time
import os
from tools.base_server import ABCToolDriver
import threading
import queue
import traceback
from typing import Optional 

if os.name == "nt":
    import clr  # type: ignore

    SOFTMAX_PRO_SDK_PATH = r"C:\Program Files (x86)\Molecular Devices\SoftMax Pro 7.1.2 Automation SDK\SoftMaxPro.AutomationClient.dll"
    clr.AddReference(SOFTMAX_PRO_SDK_PATH)  # type: ignore
    from SoftMaxPro.AutomationClient import SMPAutomationClient  # type: ignore
else:
    # Make dummy SMPAutomationClient class for non-Windows platforms to at least enable simulation mode
    import enum

    class CommandCompleted:
        def __add__(self, other: t.Any) -> "CommandCompleted":
            return self

        def __sub__(self, other: t.Any) -> "CommandCompleted":
            return self

    class InstrumentStatusChanged:
        def __add__(self, other: t.Any) -> "InstrumentStatusChanged":
            return self

        def __sub__(self, other: t.Any) -> "InstrumentStatusChanged":
            return self

    class SMPAutomationClient:  # type: ignore
        ExportAsFormat = enum.Enum("ExportAsFormat", ["COLUMNS"])

        def __init__(self) -> None:
            self.CommandCompleted = CommandCompleted()
            self.InstrumentStatusChanged = InstrumentStatusChanged()

        def Initialize(self) -> bool:
            raise NotImplementedError(
                "SMPAutomationClient is not supported on non-Windows platforms"
            )

        def Dispose(self) -> int:
            raise NotImplementedError(
                "SMPAutomationClient is not supported on non-Windows platforms"
            )

        def OpenDrawer(self) -> int:
            raise NotImplementedError(
                "SMPAutomationClient is not supported on non-Windows platforms"
            )

        def CloseDrawer(self) -> int:
            raise NotImplementedError(
                "SMPAutomationClient is not supported on non-Windows platforms"
            )

        def GetDrawerStatus(self) -> int:
            raise NotImplementedError(
                "SMPAutomationClient is not supported on non-Windows platforms"
            )

        def GetTemperature(self) -> int:
            raise NotImplementedError(
                "SMPAutomationClient is not supported on non-Windows platforms"
            )

        def GetInstrumentStatus(self) -> int:
            raise NotImplementedError(
                "SMPAutomationClient is not supported on non-Windows platforms"
            )

        def OpenFile(self, filepath: str) -> int:
            raise NotImplementedError(
                "SMPAutomationClient is not supported on non-Windows platforms"
            )

        def StartRead(self) -> int:
            raise NotImplementedError(
                "SMPAutomationClient is not supported on non-Windows platforms"
            )

        def SaveAs(self, filepath: str) -> int:
            raise NotImplementedError(
                "SMPAutomationClient is not supported on non-Windows platforms"
            )

        def ExportAs(self, filepath: str, format: str) -> int:
            raise NotImplementedError(
                "SMPAutomationClient is not supported on non-Windows platforms"
            )

        def GetDataCopy(self) -> int:
            raise NotImplementedError(
                "SMPAutomationClient is not supported on non-Windows platforms"
            )


SOFTMAX_PRO_PATH = r"C:\Program Files (x86)\Molecular Devices\SoftMax Pro 7.1.2\SoftMaxProApp.exe"
DEFAULT_PROTOCOL_DIR = r"C:\Users\Imaging Controller\Documents\spectramax_protocols"
DEFAULT_EXPERIMENT_DIR = r"C:\Users\Imaging Controller\Documents\spectramax_experiments"


class SpectramaxDriver(ABCToolDriver):
    def __init__(
        self,
        protocol_dir: str = DEFAULT_PROTOCOL_DIR,
        experiment_dir: str = DEFAULT_EXPERIMENT_DIR,
    ) -> None:
        self.protocol_dir = protocol_dir
        self.experiment_dir = experiment_dir

        self._command_lock: threading.Lock = threading.Lock()
        self.command_queue: queue.Queue[t.Any] = queue.Queue()
        self.command_response_queue: queue.Queue[t.Any] = queue.Queue()
        self.execution_thread: Optional[threading.Thread] = None
        self.live: bool = False
        self.client: SMPAutomationClient = SMPAutomationClient()
        self.softmax_process: Optional[subprocess.Popen]= None
        self.connected: bool = False
        self.instrument_status: str = "Unknown"
        self._event_lock: threading.Lock = threading.Lock()
        self.event_queue: queue.Queue[t.Any] = queue.Queue()

    def start(self) -> None:
        self.live = True
        self.execution_thread = threading.Thread(target=self.execute_spectramax_commands)
        self.execution_thread.daemon = True
        self.execution_thread.start()
        return None

    def execute_spectramax_commands(self) -> None:
        logging.info("Starting spectramax command thread")
        try:
            self.start_softmax_pro()
            while self.live:
                while not self.command_queue.empty() > 0:
                    with self._command_lock:
                        command_obj = self.command_queue.get()
                    response = self.execute_command(
                        command_obj["command"],
                        command_obj["params"],
                    )
                    event = self.wait_for_event(response)
                    with self._command_lock:
                        self.command_response_queue.put(
                            {"command": command_obj["command"], "response": event}
                        )
        except Exception:
            logging.warning("Spectramax command thread has errored")
            logging.error(traceback.format_exc())
            self.live = False
        return None

    def start_softmax_pro(self) -> None:
        if not self.connected:
            # Ignore types because STARTUPINFO is only available on Windows
            startupinfo = subprocess.STARTUPINFO()  # type: ignore
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW  # type: ignore
            self.softmax_process = subprocess.Popen(
                SOFTMAX_PRO_PATH,
                # startupinfo=startupinfo,
                #creationflags=subprocess.CREATE_NEW_CONSOLE,  # type: ignore
            )
            tries: int = 0
            while tries < 10:
                self.connected = self.client.Initialize()
                if self.connected:
                    self.client.CommandCompleted += self.handle_command_completed
                    self.client.InstrumentStatusChanged += self.handle_instrument_status_change
                    logging.info("Started softmax pro")
                    return None
                tries += 1
                time.sleep(1)
            raise Exception("Failed to start softmax pro")

    def handle_command_completed(self, sender: t.Any, event: t.Any) -> None:
        logging.info(f"Command {event.QueueID} completed")
        with self._event_lock:
            self.event_queue.put({"event_id": event.QueueID, "event": event})

    def handle_instrument_status_change(self, sender: t.Any, event: t.Any) -> None:
        logging.info(f"Status change: {event.Status}")
        self.instrument_status = event.Status

    def schedule_command(self, command: str, params: dict[str, t.Any] = {}) -> None:
        logging.info(f"Scheduling command {command}, {params}")
        with self._command_lock:
            self.command_queue.put({"command": command, "params": params})

    def execute_command(
        self, command: str, args: t.Dict[str, t.Any] = {}
    ) -> t.Any:
        command_id: int
        logging.info(f"Executing command {command}")

        if command == "open_drawer":
            command_id = self.client.OpenDrawer()
                # command_response = self.wait_for_event(command_id)
        elif command ==  "close_drawer":
                command_id = self.client.CloseDrawer()
                # command_response = self.wait_for_event(command_id)
        elif command == "get_drawer_status":
                command_id = self.client.GetDrawerStatus()
                # command_response = self.wait_for_event(command_id)
        elif command == "get_temperature":
                command_id = self.client.GetTemperature()
                # command_response = self.wait_for_event(command_id)
        elif command == "get_instrument_status":
                command_id = self.client.GetInstrumentStatus()
                # command_response = self.wait_for_event(command_id)
        elif command ==  "start_experiment":
                command_id = self.client.OpenFile(
                    os.path.join(self.protocol_dir, args["protocol_file"])
                )
                command_id = self.client.StartRead()
                command_id = self.client.SaveAs(
                    os.path.join(self.experiment_dir, args["experiment_name"] + ".sda")
                )
                command_id = self.client.ExportAs(
                    os.path.join(self.experiment_dir, args["experiment_name"] + ".txt"),
                    SMPAutomationClient.ExportAsFormat.COLUMNS,
                )
                command_id = self.client.GetDataCopy()
                # command_response = self.wait_for_event(command_id)
        else :
            raise ValueError(f"Unknown command {command}")
        return command_id
    
    def wait_for_event(self, event_id: int, timeout: Optional[int]= 60) -> t.Any:
        start_time = time.time()
        while self.live:
            seconds_spent_waiting = int(time.time() - start_time)
            if timeout and seconds_spent_waiting > timeout:
                raise Exception(
                    f"Event {event_id} has timed out. Please reset the driver."
                )

            if not self.event_queue.empty() > 0:
                with self._event_lock:
                    event_obj: dict[str, t.Any] = self.event_queue.get()
                    if event_obj["event_id"] == event_id:
                        logging.info(
                            f"Event {event_id} has completed. Waited for {seconds_spent_waiting} seconds"
                        )
                        return event_obj["event"]
                    else:
                        logging.warning(f"Unexpected command {event_obj['event_id']} received")
            if not self.live:
                break
            logging.debug(f"Time spent waiting for event_id {event_id}: {seconds_spent_waiting} seconds")
            time.sleep(1)

        if not self.live:
            raise Exception(
                "Spectramax driver has crashed. Please reset the driver"
            )
        return None

    def wait_for_command(self, command: str, timeout: Optional[int] = 60) -> t.Any:
        times = 0
        start_time = time.time()

        while self.live:
            seconds_spent_waiting = int(time.time() - start_time)
            if timeout and seconds_spent_waiting > timeout:
                raise Exception(
                    f"Command {command} has timed out. Please reset the driver."
                )

            if not self.command_response_queue.empty() > 0:
                with self._command_lock:
                    command_obj: dict[str, str] = self.command_response_queue.get()
                    if command_obj["command"] == command:
                        logging.info(
                            f"Command {command} has completed. Waited for {seconds_spent_waiting} seconds"
                        )
                        return command_obj["response"]
                    else:
                        logging.warning(f"Unexpected command {command_obj['command']} received")
            times += 1
            if not self.live:
                break

            if times % 60 == 0:
                logging.info(f"Waiting for Spectramax {command} command...({seconds_spent_waiting}s)")
            time.sleep(1)

        if not self.live:
            raise Exception(
                "Spectramax driver has crashed. Please reset the driver"
            )
        return None

    def open_drawer(self) -> None:
        self.schedule_command("open_drawer")
        self.wait_for_command('open_drawer')

    def close_drawer(self) -> None:
        self.schedule_command("close_drawer")
        self.wait_for_command('close_drawer')

    def get_drawer_status(self) -> str:
        self.schedule_command("get_drawer_status")
        event = self.wait_for_command("get_drawer_status")
        return str(event.StringResult)

    def get_temperature(self) -> float:
        self.schedule_command("get_temperature")
        event = self.wait_for_command("get_temperature")
        return float(event.DoubleResult)

    def get_instrument_status(self) -> str:
        self.schedule_command("get_instrument_status")
        event = self.wait_for_command("get_instrument_status")
        return str(event.StringResult)

    def verify_reader_communication(self) -> None:
        reader_state: str = self.get_instrument_status()

        if reader_state != "Idle":
            raise Exception(f"Expected reader state Idle. Got reader state: {reader_state}")

    def start_experiment(
        self, experiment_name: str, protocol_file: str = "96wellblk_OD600_A1.spr"
    ) -> str:
        self.schedule_command(
            command="start_experiment",
            params={"experiment_name": experiment_name, "protocol_file": protocol_file},
        )
        event = self.wait_for_command("start_experiment")
        return str(event.StringResult)

    def close(self) -> None:
        self.live = False
        if self.execution_thread:
            self.execution_thread.join()
        self.client.CommandCompleted -= self.handle_command_completed
        self.client.InstrumentStatusChanged -= self.handle_instrument_status_change
        self.client.Dispose()
        if self.softmax_process:
            self.softmax_process.terminate()
            self.softmax_process.wait()
            self.softmax_process = None


if __name__ == "__main__":
    sm = SpectramaxDriver()
    try:
        print("Status: ", sm.instrument_status)
        sm.start()
        print(sm.get_instrument_status())
        print(sm.get_temperature())
        print(sm.get_drawer_status())
        sm.open_drawer()
        sm.close_drawer()
    finally:
        sm.close()


import logging

from tools.base_server import ToolServer, serve
from tools.grpc_interfaces.spectramax_pb2 import Command, Config
import sys
from .driver import SpectramaxDriver
import argparse

class SpectramaxServer(ToolServer):
    toolType = "spectramax"

    def __init__(self) -> None:
        super().__init__()
        self.driver: SpectramaxDriver

    def _configure(self, config: Config) -> None:
        self.config = config
        if self.driver:
            self.driver.close()
        self.driver = SpectramaxDriver(
            protocol_dir=config.protocol_dir, experiment_dir=config.experiment_dir
        )
        self.driver.start()
        self.driver.verify_reader_communication()

    def OpenDrawer(self, params: Command.OpenDrawer) -> None:
        self.driver.open_drawer()

    def CloseDrawer(self, params: Command.CloseDrawer) -> None:
        self.driver.close_drawer()

    def StartRead(self, params: Command.StartRead) -> None:
        self.driver.start_experiment(
            protocol_file=params.protocol_file, experiment_name=params.experiment_name
        )

    def EstimateOpenDrawer(self, params: Command.OpenDrawer) -> int:
        return 2

    def EstimateCloseDrawer(self, params: Command.CloseDrawer) -> int:
        return 2

    def EstimateStartRead(self, params: Command.StartRead) -> int:
        return 5


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument('--port')
    known, remaining = parser.parse_known_args()
    sys.argv = [sys.argv[0]] + remaining
    
    if not known.port:
        raise RuntimeWarning("Port must be provided...")
    serve(SpectramaxServer(),str(known.port))




import os
import logging 
from tools.base_server import ABCToolDriver
import subprocess 
import time 

if os.name == "nt":
    import clr  # type: ignore

    SDK_NAME = "HiGIntegration.dll"
    common_paths = ["C:\\HiG\\", "C:\\Program Files (x86)\\BioNex\\HiG\\"]
    for path in common_paths:
        full_path :str = os.path.join(path,SDK_NAME)
        if os.path.exists(full_path):
            clr.AddReference(full_path) 
            from BioNex import HiGIntegration # type: ignore
            break

else:
    class HiGIntegration: #type: ignore
        def __init__(self, can_port:int)-> None:
            self.can_port: int = can_port
        
        def Initialize(self, name:str, can_port:str, blocking:bool) -> None:
            raise NotImplementedError(
                "HiGInterface is not supported on non-Windows platforms"
            )
        def Home(self) -> None:
            raise NotImplementedError(
                "HiGInterface is not supported on non-Windows platforms"
            )
        def Close(self) -> None:
            raise NotImplementedError(
                "HiGInterface is not supported on non-Windows platforms"
            )
        def ShowDiagnostics(self) -> None:
            raise NotImplementedError(
                "HiGInterface is not supported on non-Windows platforms"
            )
        def OpenShield(self, bucket:int) -> None:
            raise NotImplementedError(
                "HiGInterface is not supported on non-Windows platforms"
            )
        def CloseShield(self) -> None:
            raise NotImplementedError(
                "HiGInterface is not supported on non-Windows platforms"
            )
        def Spin(self,speed_g:int, accel_percent:int,  decel_percent:int, time_seconds:int) -> None:
            raise NotImplementedError(
                "HiGInterface is not supported on non-Windows platforms"
            )
        def AbortSpin(self) -> None:
            raise NotImplementedError(
                "HiGInterface is not supported on non-Windows platforms"
            )
        def HomeShield(self) -> None:
            raise NotImplementedError(
                "HiGInterface is not supported on non-Windows platforms"
            )
        def Dispose(self) -> None:
             raise NotImplementedError(
                "HiGInterface is not supported on non-Windows platforms"
            )           

class HiGCentrifugeDriver(ABCToolDriver):

    def __init__(self , can_port:int) -> None:
        self.can_port: int = can_port
        self.live : bool  = False
        self.client : HiGIntegration
        self.initialize()

    def initialize(self) -> None:
        try:
            p = subprocess.Popen('C:\Program Files (x86)\BioNex\HiG\CanDongleServerProcess.exe', stdout=subprocess.PIPE,shell=True)
            p.poll()
            time.sleep(0.5)
            self.client = HiGIntegration.HiG() #ignore
            self.client.Blocking = True #ignore
            self.client.Initialize("HiG1", str(self.can_port),False) #ignore
            self.live = True 

        except RuntimeError:
            logging.debug("Failed to initialize HiG") 

    def home(self)-> None:
        self.client.Home()

    def spin(self, speed_g:int, accel_percent:int,  decel_percent:int, time_seconds:int)-> None:
        self.client.Spin(speed_g, accel_percent, decel_percent, time_seconds)
    
    def close_shield(self)-> None:
        self.client.CloseShield()
    
    def open_shield(self, bucket:int) -> None:
        self.client.OpenShield(bucket)
    
    def close(self) -> None:
        self.live = False
        self.client.Close()
    
    def show_diagnostics(self) -> None:
        self.client.ShowDiagnostics() 

    def abort_spin(self) -> None:
        return
    


import json
import logging
from pathlib import Path
from typing import Any, Union

import appdirs  # type: ignore
import requests

APP_NAME = "galago"
APP_AUTHOR = "sciencecorp"
DATA_DIR = Path(appdirs.user_data_dir(APP_NAME, APP_AUTHOR))
CONFIG_FILE = DATA_DIR / "api_config.json"


class Db:
    _api_url = "http://localhost:8000/api"  # Default

    @classmethod
    def get_api_url(cls) -> str:
        """Get the current API URL from config file or use default"""
        try:
            if CONFIG_FILE.exists():
                with open(CONFIG_FILE, "r") as f:
                    config: dict[str, Any] = json.load(f)
                    url = config.get("api_url", cls._api_url)
                    return str(url) if url else cls._api_url
        except Exception as e:
            logging.warning(f"Failed to read API config: {e}")
        return cls._api_url

    @classmethod
    def set_api_url(cls, url: str) -> bool:
        """Save API URL to config file"""
        try:
            CONFIG_FILE.parent.mkdir(parents=True, exist_ok=True)
            with open(CONFIG_FILE, "w") as f:
                json.dump({"api_url": url}, f)
            cls._api_url = url
            logging.info(f"API URL updated to: {url}")
            return True
        except Exception as e:
            logging.error(f"Failed to save API config: {e}")
            return False

    @classmethod
    def check_connection(cls) -> bool:
        try:
            api_url = cls.get_api_url()
            response = requests.get(api_url, timeout=5)
            if response.status_code == 200:
                return True
        except requests.exceptions.ConnectionError:
            return False
        except requests.exceptions.Timeout:
            return False
        return False

    @classmethod
    def get_data(cls, model: str) -> Any:
        api_url = cls.get_api_url()
        response = requests.get(f"{api_url}/{model}")
        return response.json()

    @classmethod
    def get_by_id_or_name(cls, id: Union[int, str], model: str) -> Any:
        api_url = cls.get_api_url()
        response = requests.get(f"{api_url}/{model}/{id}")
        if response.status_code == 404:
            logging.warning(f"Resource with id/name {id} not found in {model}.")
            return None
        return response.json()

    @classmethod
    def post_data(cls, data: dict, model: str) -> Any:
        api_url = cls.get_api_url()
        response = requests.post(f"{api_url}/{model}", json=data)
        return response.json()

    @classmethod
    def delete_data(cls, id: Union[int, str], model: str) -> Any:
        api_url = cls.get_api_url()
        response = requests.delete(f"{api_url}/{model}/{id}")
        return response.json()

    @classmethod
    def update_data(cls, id: Union[str, int], data: dict, model: str) -> Any:
        api_url = cls.get_api_url()
        response = requests.put(f"{api_url}/{model}/{id}", json=data)
        return response.json()

    @classmethod
    def ping(cls, times: int) -> bool:
        api_url = cls.get_api_url()
        for i in range(times):
            try:
                response = requests.get(f"{api_url}/health", timeout=5)
                logging.info(f"Response status code: {response.status_code}")
                if response.status_code == 200:
                    return True
            except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
                continue
        return False



import logging

from tools.base_server import ToolServer, serve
from tools.grpc_interfaces.dataman70_pb2 import Command, Config

from .driver import Dataman70Driver
import argparse

class Dataman70Server(ToolServer):
  toolType = "dataman70"
  driver: Dataman70Driver
  config: Config

  def __init__(self) -> None:
    super().__init__()
  
  def _configure(self, config: Config) -> None:
    self.config = config
    if self.driver:
      self.driver.close()
    self.driver = Dataman70Driver(self.config.com_port)
  
  def Scan(self, params: Command.Scan) -> None:
    self.driver.scan_barcode(params.mapped_variable)
  
  def AssertBarcode(self, params: Command.AssertBarcode) -> None:
    self.driver.assert_barcode(params.barcode)

  def EstimateScan(self, params: Command.Scan) -> int:
    return 5
  
  def EstimateAssertBarcode(self, params: Command.AssertBarcode) -> int:
    return 5

if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument('--port')
    args = parser.parse_args()
    if not args.port:
        raise RuntimeWarning("Port must be provided...")
    serve(Dataman70Server(), str(args.port))



import argparse
import asyncio
import json
import logging
import logging.handlers
import os
import signal as os_signal
import socket
import subprocess
import sys
import threading
import time
import webbrowser
from http.server import HTTPServer, SimpleHTTPRequestHandler
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple

import appdirs  # type: ignore
import requests
import websockets
from colorama import Fore, Style, init
from packaging import version

from tools import __version__ as galago_version
from tools.app_config import Config
from tools.toolbox.db import Db
from tools.utils import get_local_ip, get_shell_command

# Add the project root to Python path
ROOT_DIR = Path(__file__).parent
sys.path.insert(0, str(ROOT_DIR))

# Configuration flags - matching ToolsManager
USE_APP_DATA_DIR = True  # Set to False for local development/testing

# Use appdirs to get platform-specific data directory
APP_NAME = "galago"
APP_AUTHOR = "sciencecorp"
DATA_DIR = appdirs.user_data_dir(APP_NAME, APP_AUTHOR)

LOG_TIME = int(time.time())
LOCAL_IP = get_local_ip()


init(autoreset=True)


def setup_logging() -> Path:
    """Setup logging similar to ToolsManager"""
    # Set appropriate log directory based on configuration
    if USE_APP_DATA_DIR:
        # Use platform-specific app data directory (for production)
        log_folder = Path(DATA_DIR) / "trace_logs" / str(LOG_TIME)
    else:
        # Use directory relative to ROOT_DIR (for local development/testing)
        log_folder = ROOT_DIR / "data" / "trace_logs" / str(LOG_TIME)

    # Ensure the log directory exists
    try:
        log_folder.mkdir(parents=True, exist_ok=True)
    except Exception as e:
        print(f"Failed to create log directory: {log_folder}. Error: {str(e)}")
        log_folder = Path.home() / "galago_logs" / str(LOG_TIME)
        log_folder.mkdir(parents=True, exist_ok=True)

    # Setup logging configuration
    log_file = log_folder / "web_server.log"

    # Create formatters
    formatter = logging.Formatter(
        "%(asctime)s | %(levelname)s | %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
    )

    root_logger = logging.getLogger()
    root_logger.handlers.clear()

    # Setup file handler with rotation
    file_handler = logging.handlers.RotatingFileHandler(
        log_file, maxBytes=10 * 1024 * 1024, backupCount=5
    )
    file_handler.setFormatter(formatter)
    file_handler.setLevel(logging.INFO)

    # Setup console handler - only INFO and above
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(formatter)
    console_handler.setLevel(logging.INFO)

    # Configure root logger
    root_logger.setLevel(logging.INFO)
    root_logger.addHandler(file_handler)
    root_logger.addHandler(console_handler)

    return log_folder


def check_for_updates() -> Tuple[bool, str, str]:
    """Check if there's a newer version of galago-tools on PyPI"""
    try:
        current_version = galago_version
        response = requests.get("https://pypi.org/pypi/galago-tools/json", timeout=3)
        if response.status_code == 200:
            data = response.json()
            latest_version = data["info"]["version"]

            # Compare versions using packaging.version for proper semantic versioning comparison
            if version.parse(latest_version) > version.parse(current_version):
                return True, current_version, latest_version
            else:
                return False, current_version, latest_version
    except Exception as e:
        logging.warning(f"Failed to check for updates: {str(e)}")

    # Return default values if the check fails
    return False, galago_version, galago_version


def display_startup_message(
    log_folder: Path,
    update_available: bool = False,
    current_version: str = "",
    latest_version: str = "",
) -> None:
    current_time = time.strftime("%Y-%m-%d %H:%M:%S")

    # Detect if we can safely use emojis
    try:
        # Test if stdout can handle emojis
        sys.stdout.write("\U0001f680")
        sys.stdout.flush()
        use_emojis = True
    except (UnicodeEncodeError, UnicodeDecodeError):
        use_emojis = False

    # Choose symbols based on encoding support
    if use_emojis:
        rocket = "🚀"
        robot = "🤖"
        package = "📦"
        clock = "⏰"
        id_icon = "🆔"
        computer = "💻"
        folder = "📂"
        checkmark = "✅"
        refresh = "🔄"
    else:
        rocket = "*"
        robot = "*"
        package = "[*]"
        clock = "[T]"
        id_icon = "[#]"
        computer = "[PC]"
        folder = "[DIR]"
        checkmark = "[OK]"
        refresh = "[>>]"

    print(f"{Fore.CYAN}{'=' * 80}{Style.RESET_ALL}")
    print(f"{Fore.GREEN}{Style.BRIGHT}{rocket} {robot} GALAGO WEB SERVER STARTED{Style.RESET_ALL}")
    print(f"{Fore.CYAN}{'=' * 80}{Style.RESET_ALL}")
    print("")
    print(f"{Fore.YELLOW}{package} Version:{Style.RESET_ALL} {galago_version}")
    if update_available:
        print(f"{Fore.RED}    A new version ({latest_version}) is available!{Style.RESET_ALL}")
        print(
            f"{Fore.YELLOW}    Upgrade using: pip install --upgrade galago-tools{Style.RESET_ALL}"
        )
    print(f"{Fore.YELLOW}{clock} Started:{Style.RESET_ALL} {current_time}")
    print(f"{Fore.YELLOW}{id_icon} Session:{Style.RESET_ALL} {LOG_TIME}")
    print(f"{Fore.YELLOW}{computer} Platform:{Style.RESET_ALL} {os.name}")
    print("")
    print(f"{Fore.MAGENTA}{folder} URLs:{Style.RESET_ALL}")
    print(f"{Fore.CYAN}   Tool Server IP: {LOCAL_IP}{Style.RESET_ALL}")
    print(f"{Fore.CYAN}   Web Interface Local: http://localhost:8080/{Style.RESET_ALL}")
    print(f"{Fore.CYAN}   Web Interface On Network: http://{LOCAL_IP}:8080/{Style.RESET_ALL}")
    print(f"{Fore.CYAN}   Logs Directory: {log_folder}{Style.RESET_ALL}")
    print("")
    print(f"{Fore.GREEN}{checkmark} Web Server initialized successfully{Style.RESET_ALL}")
    print(f"{Fore.BLUE}{refresh} Starting WebSocket and HTTP servers...{Style.RESET_ALL}")
    print(f"{Fore.CYAN}{'-' * 80}{Style.RESET_ALL}")
    print("")


# Global state
connected_clients: Set[Any] = set()
server_processes: Dict[str, subprocess.Popen] = {}
config: Optional[Config] = None
log_folder: Optional[Path] = None
log_positions: Dict[str, int] = {}
last_tool_status: Dict[str, str] = {}
logger = logging.getLogger(__name__)


def is_process_running(tool_name: str) -> bool:
    """Check if a process is running"""
    if tool_name not in server_processes:
        return False
    process = server_processes[tool_name]
    return process is not None and process.poll() is None


def is_port_occupied(port: int) -> bool:
    """Check if a port is occupied"""
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            result = sock.connect_ex(("127.0.0.1", port))
            return result == 0
    except Exception:
        return False


def open_browser(url: str, delay: float = 2.0) -> None:
    """Open browser after a delay to ensure server is ready"""

    def delayed_open() -> None:
        time.sleep(delay)
        try:
            # Try to verify the server is responding before opening browser
            max_attempts = 10
            for attempt in range(max_attempts):
                try:
                    import urllib.request

                    urllib.request.urlopen(url, timeout=1)
                    break
                except Exception as e:
                    if attempt < max_attempts - 1:
                        time.sleep(0.5)
                    else:
                        logger.warning(
                            f"Server not responding at {url} after {max_attempts} attempts: {e}"
                        )

            webbrowser.open(url)
        except Exception as e:
            logger.error(f"Failed to open browser: {e}")
            logger.info(f"Please manually open: {url}")

    thread = threading.Thread(target=delayed_open, daemon=True)
    thread.start()


def kill_by_process_id(process_id: int) -> None:
    """Kill a process by PID"""
    try:
        if os.name == "nt":
            subprocess.call(["taskkill", "/F", "/T", "/PID", str(process_id)])
        else:
            os.kill(process_id, os_signal.SIGTERM)
            time.sleep(0.5)
            try:
                os.kill(process_id, os_signal.SIGINT)
            except ProcessLookupError:
                pass
        logger.info(f"Killed process {process_id}")
    except Exception as e:
        logger.error(f"Failed to kill process {process_id}: {e}")


def kill_process_by_name(process_name: str) -> None:
    """Kill a process by name"""
    if process_name not in server_processes:
        return
    try:
        process = server_processes[process_name]
        kill_by_process_id(process.pid)
        del server_processes[process_name]
        logger.info(f"Killed process {process_name}")
    except Exception as e:
        logger.warning(f"Failed to kill process {process_name}: {e}")


async def get_tool_status() -> List[Dict[str, Any]]:
    """Get current status of all tools"""
    tools_status: List[Dict[str, Any]] = []

    # Add toolbox
    toolbox_running = is_process_running("Tool Box")
    tools_status.append(
        {
            "name": "Tool Box",
            "type": "toolbox",
            "port": 1010,
            "status": "running" if toolbox_running else "stopped",
            "image": "toolbox.png",
        }
    )

    # Add workcell tools
    if config and config.workcell_config:
        for tool in config.workcell_config.tools:
            tool_running = is_process_running(tool.name)
            tools_status.append(
                {
                    "name": tool.name,
                    "type": tool.type,
                    "port": tool.port,
                    "status": "running" if tool_running else "stopped",
                    "image": f"{tool.type}.png",
                }
            )

    return tools_status


async def reload_config() -> bool:
    """Reload the configuration from disk"""
    global config, last_tool_status

    try:
        logger.info("Reloading configuration...")
        config = Config()
        config.load_workcell_config()

        # Reset status tracking to force update
        last_tool_status = {}

        logger.info("Configuration reloaded successfully")
        return True

    except Exception as e:
        logger.error(f"Failed to reload configuration: {e}")
        return False


async def relaunch_all_tools() -> bool:
    """Stop all tools, reload config, then start all tools"""
    try:
        logger.info("Starting tool relaunch sequence...")

        # Step 1: Stop all running tools
        current_tools = await get_tool_status()
        running_tools = [tool for tool in current_tools if tool["status"] == "running"]

        if running_tools:
            logger.info(f"Stopping {len(running_tools)} running tools...")
            for tool in running_tools:
                await stop_tool(tool["name"])

            # Wait for all tools to stop
            await asyncio.sleep(2)

        # Step 2: Reload configuration
        config_success = await reload_config()
        if not config_success:
            raise Exception("Failed to reload configuration")

        # Step 3: Start all tools from new config
        logger.info("Starting all tools with new configuration...")
        new_tools = await get_tool_status()

        start_tasks = []
        for tool in new_tools:
            if tool["name"] != "Tool Box":  # Skip toolbox if you don't want to auto-start it
                start_tasks.append(start_tool(tool["name"], tool["type"], tool["port"]))

        # Start all tools concurrently but with small delays
        for i, task in enumerate(start_tasks):
            await asyncio.sleep(0.5 * i)  # Stagger starts
            asyncio.create_task(task)

        logger.info("Tool relaunch sequence completed")
        return True

    except Exception as e:
        logger.error(f"Failed to relaunch tools: {e}")
        return False


async def check_for_status_changes() -> None:
    """Check if tool status has changed and broadcast updates"""
    global last_tool_status

    current_status = await get_tool_status()

    # Convert to dict for easier comparison
    current_status_dict = {tool["name"]: tool["status"] for tool in current_status}

    # Check if status has changed
    if current_status_dict != last_tool_status:
        await send_tool_status()
        last_tool_status = current_status_dict.copy()


async def monitor_tool_processes() -> None:
    """Monitor tool processes and broadcast status changes"""
    while True:
        try:
            await check_for_status_changes()
        except Exception as e:
            logger.error(f"Error in process monitoring: {e}")

        await asyncio.sleep(2)  # Check every 2 seconds


async def start_tool(tool_name: str, tool_type: str, port: int) -> bool:
    """Start a tool process with improved logging"""
    try:
        kill_process_by_name(tool_name)

        if is_port_occupied(port):
            raise Exception(f"Port {port} is already occupied")

        cmd = get_shell_command(tool_type=tool_type, port=port)
        os.chdir(ROOT_DIR)

        if log_folder:
            log_file = log_folder / f"{tool_name}.log"

            # Initialize log position tracking before starting the process
            log_positions[str(log_file)] = 0

            # Create the log file first and add an initial entry
            with open(log_file, "w") as f:
                f.write(f"Starting {tool_name} ({tool_type}) on port {port}\n")
                f.flush()  # Force immediate write

            # Open log file for subprocess with line buffering
            log_handle = open(log_file, "a", buffering=1)  # Line buffered

            process = subprocess.Popen(
                cmd,
                stdout=log_handle,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
                shell=os.name == "nt",
                bufsize=1,  # Line buffered
            )

            # Store the file handle so we can close it later
            if not hasattr(process, "_log_handles"):
                setattr(process, "_log_handles", [])
            getattr(process, "_log_handles").append(log_handle)

        else:
            process = subprocess.Popen(cmd, universal_newlines=True, shell=os.name == "nt")

        server_processes[tool_name] = process
        logger.info(f"Started {tool_name} on port {port}")

        # Immediately check for new logs after a brief delay
        asyncio.create_task(immediate_log_check(tool_name))

        # Schedule a status check after a brief delay to catch the change
        asyncio.create_task(delayed_status_check())

        return True

    except Exception as e:
        logger.error(f"Failed to start {tool_name}: {e}")
        return False


async def stop_tool(tool_name: str) -> bool:
    """Stop a tool process with proper cleanup"""
    try:
        if tool_name in server_processes:
            process = server_processes[tool_name]

            # Close any open log file handles
            if hasattr(process, "_log_handles"):
                for handle in process._log_handles:
                    try:
                        handle.close()
                    except Exception as e:
                        logger.warning(f"Error closing log file {handle.name}: {e}")

        kill_process_by_name(tool_name)
        logger.info(f"Stopped {tool_name}")

        # Schedule a status check after a brief delay to catch the change
        asyncio.create_task(delayed_status_check())

        return True
    except Exception as e:
        logger.error(f"Failed to stop {tool_name}: {e}")
        return False


async def delayed_status_check() -> None:
    """Check status after a brief delay"""
    await asyncio.sleep(1)
    await check_for_status_changes()


async def get_recent_logs(lines: int = 100) -> List[Dict[str, Any]]:
    """Get recent logs from all log files"""
    logs: List[Dict[str, Any]] = []

    if not log_folder:
        return logs

    try:
        for log_file_path in log_folder.glob("*.log"):
            try:
                with open(log_file_path, "r", encoding="utf-8") as f:
                    file_lines = f.readlines()
                    recent_lines = file_lines[-lines:] if len(file_lines) > lines else file_lines

                    for line in recent_lines:
                        logs.append(
                            {
                                "source": log_file_path.stem,
                                "content": line.strip(),
                                "timestamp": time.time(),
                            }
                        )
            except Exception as e:
                logger.error(f"Error reading log file {log_file_path}: {e}")
    except Exception as e:
        logger.error(f"Error accessing log folder: {e}")

    logs.sort(key=lambda x: x["timestamp"], reverse=True)
    return logs[:lines]


async def immediate_log_check(tool_name: str, delay: float = 0.1) -> None:
    """Immediately check for logs from a newly started tool"""
    if not log_folder:
        return

    log_file = log_folder / f"{tool_name}.log"

    # Wait a bit for initial logs to be written
    for i in range(10):  # Check up to 10 times over 1 second
        await asyncio.sleep(delay)

        if log_file.exists():
            try:
                with open(log_file, "r", encoding="utf-8") as f:
                    content = f.read().strip()
                    if content:
                        lines = content.split("\n")
                        logs = []
                        for line in lines:
                            if line.strip():
                                logs.append(
                                    {
                                        "source": tool_name,
                                        "content": line.strip(),
                                        "timestamp": time.time(),
                                    }
                                )

                        if logs:
                            message = {"type": "logs", "data": logs}
                            await broadcast_message(message)

                            # Update position tracker
                            log_positions[str(log_file)] = log_file.stat().st_size
                            break

            except Exception as e:
                logger.error(f"Error in immediate log check for {tool_name}: {e}")


async def broadcast_message(message: Dict[str, Any]) -> None:
    """Send message to all connected clients"""
    if not connected_clients:
        return

    disconnected = set()
    for client in connected_clients.copy():
        try:
            await client.send(json.dumps(message))
        except websockets.exceptions.ConnectionClosed:
            disconnected.add(client)
        except Exception as e:
            logger.error(f"Error broadcasting to client: {e}")
            disconnected.add(client)

    for client in disconnected:
        connected_clients.discard(client)


async def send_tool_status(websocket: Optional[Any] = None) -> None:
    """Send current tool status to client(s)"""
    tools_status = await get_tool_status()

    message = {"type": "tool_status", "data": tools_status}

    if websocket:
        try:
            await websocket.send(json.dumps(message))
        except websockets.exceptions.ConnectionClosed:
            pass
    else:
        await broadcast_message(message)


async def handle_websocket_message(websocket: Any, data: Dict[str, Any]) -> None:
    """Handle incoming WebSocket messages"""
    action = data.get("action")
    tool_name = data.get("tool_name")
    tool_type = data.get("tool_type")
    port_raw = data.get("port")

    # Convert port to int safely
    port = 0
    if port_raw is not None:
        try:
            port = int(port_raw)
        except (ValueError, TypeError):
            port = 0

    response: Dict[str, Any] = {"type": "response", "success": False, "message": "Unknown action"}

    try:
        if action == "start_tool":
            success = await start_tool(str(tool_name), str(tool_type), int(port))
            response = {
                "type": "response",
                "success": success,
                "message": f"{'Started' if success else 'Failed to start'} {tool_name}",
            }

        elif action == "stop_tool":
            success = await stop_tool(str(tool_name))
            response = {
                "type": "response",
                "success": success,
                "message": f"{'Stopped' if success else 'Failed to stop'} {tool_name}",
            }

        elif action == "reload_config":
            success = await reload_config()
            response = {
                "type": "response",
                "success": success,
                "message": f"{'Configuration reloaded successfully' if success else 'Failed to reload configuration'}",
            }
            if success:
                # Force a status update to show any new tools
                await send_tool_status()

        elif action == "relaunch_all":
            success = await relaunch_all_tools()
            response = {
                "type": "response",
                "success": success,
                "message": f"{'All tools relaunched successfully' if success else 'Failed to relaunch tools'}",
            }
        elif action == "get_api_url":
            current_url = Db.get_api_url()
            response = {"type": "api_url", "url": current_url, "success": True}
        elif action == "set_api_url":
            new_url = data.get("url", "").strip()
            if not new_url:
                response = {
                    "type": "response",
                    "success": False,
                    "message": "API URL cannot be empty",
                }
            else:
                # Validate URL format
                if not new_url.startswith(("http://", "https://")):
                    response = {
                        "type": "response",
                        "success": False,
                        "message": "API URL must start with http:// or https://",
                    }
                else:
                    success = Db.set_api_url(new_url)
                    response = {
                        "type": "response",
                        "success": success,
                        "message": f"{'API URL updated successfully to: ' + new_url if success else 'Failed to update API URL'}",
                    }
                    if success:
                        # Reload config with new API URL
                        await reload_config()

        elif action == "get_logs":
            logs = await get_recent_logs()
            response = {"type": "logs", "data": logs}

        elif action == "get_status":
            await send_tool_status(websocket)
            return

        await websocket.send(json.dumps(response))

    except Exception as e:
        logger.error(f"Error handling action {action}: {e}")
        error_response = {
            "type": "response",
            "success": False,
            "message": f"Error processing {action}: {str(e)}",
        }
        await websocket.send(json.dumps(error_response))


async def monitor_log_files() -> None:
    """Enhanced log file monitoring with better error handling"""

    while True:
        try:
            if not log_folder or not connected_clients:
                await asyncio.sleep(0.5)
                continue

            for log_file in log_folder.glob("*.log"):
                # Skip monitoring the web server's own log file to prevent feedback loop
                if log_file.stem == "web_server":
                    continue

                file_path = str(log_file)

                try:
                    current_size = log_file.stat().st_size
                    last_position = log_positions.get(file_path, 0)

                    if current_size > last_position:
                        with open(log_file, "r", encoding="utf-8") as f:
                            f.seek(last_position)
                            new_content = f.read()

                            if new_content.strip():
                                lines = new_content.strip().split("\n")
                                logs = []
                                for line in lines:
                                    if line.strip():
                                        logs.append(
                                            {
                                                "source": log_file.stem,
                                                "content": line.strip(),
                                                "timestamp": time.time(),
                                            }
                                        )

                                if logs:
                                    message = {"type": "logs", "data": logs}
                                    await broadcast_message(message)

                            log_positions[file_path] = current_size

                    elif current_size < last_position:
                        # File was truncated or recreated, reset position
                        log_positions[file_path] = 0

                except FileNotFoundError:
                    # Log file was deleted, remove from tracking
                    if file_path in log_positions:
                        del log_positions[file_path]
                except Exception as e:
                    logger.error(f"Error monitoring log file {log_file}: {e}")

        except Exception as e:
            logger.error(f"Error in log monitoring: {e}")

        await asyncio.sleep(0.2)


async def websocket_handler(websocket: Any) -> None:
    """Handle WebSocket connections"""
    connected_clients.add(websocket)
    logger.info(f"Client connected. Total: {len(connected_clients)}")

    try:
        await send_tool_status(websocket)

        # Send initial logs with different message type
        logs = await get_recent_logs(50)
        if logs:
            initial_message = {
                "type": "initial_logs",  # Use different type for initial logs
                "data": logs,
            }
            await websocket.send(json.dumps(initial_message))

        async for message in websocket:
            try:
                data = json.loads(message)
                await handle_websocket_message(websocket, data)
            except json.JSONDecodeError:
                error_response = {
                    "type": "response",
                    "success": False,
                    "message": "Invalid JSON message",
                }
                await websocket.send(json.dumps(error_response))
            except Exception as e:
                logger.error(f"Error handling WebSocket message: {e}")

    except websockets.exceptions.ConnectionClosed:
        logger.info("Client connection closed")
    except Exception as e:
        logger.error(f"WebSocket error: {e}")
    finally:
        connected_clients.discard(websocket)


class CustomHTTPRequestHandler(SimpleHTTPRequestHandler):
    """Custom HTTP handler to serve static files"""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, directory=str(ROOT_DIR), **kwargs)

    def log_message(self, format: str, *args: Any) -> None:
        if args[1] != "200":
            super().log_message(format, *args)

    def end_headers(self) -> None:
        self.send_header("Access-Control-Allow-Origin", "*")
        self.send_header("Access-Control-Allow-Methods", "GET, POST, OPTIONS")
        self.send_header("Access-Control-Allow-Headers", "Content-Type")
        super().end_headers()

    def do_GET(self) -> None:
        if self.path == "/" or self.path == "/index.html":
            html_file = ROOT_DIR / "index.html"

            if html_file.exists():
                self.send_response(200)
                self.send_header("Content-type", "text/html")
                self.send_header("Cache-Control", "no-cache")
                self.end_headers()
                with open(html_file, "rb") as f:
                    self.wfile.write(f.read())
                return
            else:
                self.send_response(404)
                self.send_header("Content-type", "text/html")
                self.end_headers()
                error_html = f"""
                <!DOCTYPE html>
                <html>
                <body>
                    <h1>404 - index.html not found</h1>
                    <p>Make sure index.html is in the project root directory.</p>
                    <p>Current directory: {ROOT_DIR}</p>
                </body>
                </html>
                """
                self.wfile.write(error_html.encode("utf-8"))
                return

        if self.path.startswith("/images/"):
            image_name = self.path[8:]

            images_path = ROOT_DIR / "tool_images" / image_name

            if images_path.exists():
                self.send_response(200)

                if image_name.lower().endswith(".png"):
                    self.send_header("Content-type", "image/png")
                elif image_name.lower().endswith((".jpg", ".jpeg")):
                    self.send_header("Content-type", "image/jpeg")
                elif image_name.lower().endswith(".gif"):
                    self.send_header("Content-type", "image/gif")
                elif image_name.lower().endswith(".svg"):
                    self.send_header("Content-type", "image/svg+xml")
                else:
                    self.send_header("Content-type", "image/png")

                self.end_headers()
                with open(images_path, "rb") as f:
                    self.wfile.write(f.read())
                return

            logger.warning(f"Image not found: {image_name}")
            self.send_response(404)
            self.end_headers()
            return

        super().do_GET()


def run_http_server(port: int = 8080) -> None:
    """Run HTTP server in a separate thread"""
    server = HTTPServer(("localhost", port), CustomHTTPRequestHandler)
    server.serve_forever()


def cleanup_processes() -> None:
    """Cleanup all processes"""
    logger.info("Cleaning up processes...")
    for tool_name in list(server_processes.keys()):
        kill_process_by_name(tool_name)


def parse_arguments() -> argparse.Namespace:
    """Parse command line arguments"""
    parser = argparse.ArgumentParser(description="Galago Web Server")
    parser.add_argument(
        "--api-url",
        type=str,
        default=None,
        help="API URL for the Galago backend (default: http://localhost:8000/api)",
    )

    return parser.parse_args()


async def main() -> None:
    """Main function"""
    global config, log_folder, last_tool_status

    try:
        # Parse command line arguments
        args = parse_arguments()

        # Setup logging first
        log_folder = setup_logging()

        # Set API URL if provided via command line
        if args.api_url:
            logger.info(f"Setting API URL from command line: {args.api_url}")
            Db.set_api_url(args.api_url)
        else:
            # Load from config or use default
            current_url = Db.get_api_url()
            logger.info(f"Using API URL: {current_url}")

        # Check for updates
        update_available, current_version, latest_version = check_for_updates()

        # Display startup message
        display_startup_message(log_folder, update_available, current_version, latest_version)

        # Initialize config
        config = Config()
        config.load_workcell_config()

        # Initialize last_tool_status
        initial_status = await get_tool_status()
        last_tool_status = {tool["name"]: tool["status"] for tool in initial_status}

        # Start HTTP server in background thread
        http_port = 8080
        http_thread = threading.Thread(target=run_http_server, args=(http_port,), daemon=True)
        http_thread.start()

        # Start WebSocket server
        server = await websockets.serve(websocket_handler, "localhost", 8765)

        # Start monitoring tasks
        asyncio.create_task(monitor_log_files())
        asyncio.create_task(monitor_tool_processes())

        # open_browser(browser_url, delay=2.0)
        # Wait for server to close
        await server.wait_closed()

    except Exception as e:
        logger.error(f"Failed to start server: {e}")
        return
    finally:
        cleanup_processes()


if __name__ == "__main__":
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        logger.info("Server stopped by user")
        print("\n🛑 Galago Web Server stopped by user")
        print("=" * 80)



import logging
from tools.base_server import ToolServer, serve
from tools.grpc_interfaces.plateloc_pb2 import Command, Config
from .driver import PlateLocDriver
import argparse 

class PlateLocServer(ToolServer):
     toolType = "plateloc"

     def __init__(self) -> None:
          super().__init__()
          self.driver : PlateLocDriver

     def _configure(self, request:Config) -> None:
          logging.info("Initializing Plateloc...")
          self.config = request
          self.driver = PlateLocDriver(request.profile)
     
     def ShowDiagsDialog(self, params:Command.ShowDiagsDialog) -> None:
          self.driver.show_diagnostics()
     
     def SetTemperature(self, params:Command.SetTemperature) -> None:
          logging.info(f"Setting temperature to {params.temperature}")
          self.driver.set_temperature(params.temperature)
     
     def SetSealTime(self, params:Command.SetSealTime) -> None:
          logging.info(f"Setting time to {params.time}")
          self.driver.set_seal_time(params.time)
     
     def GetActualTemperature(self, params:Command.GetActualTemperature) -> None:
          self.driver.get_actual_temperature()
     
     def StageIn(self, params:Command.StageIn) -> None:
          self.driver.stage_in()
     
     def StageOut(self, params:Command.StageOut) -> None:
          self.driver.stage_out()
     
     def Seal(self, params:Command.Seal) -> None:
          self.driver.seal()
          self.driver.stage_out()
     

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument('--port')
    args = parser.parse_args()
    if not args.port:
         raise RuntimeWarning("Port must be provided...")
    serve(PlateLocServer(), str(args.port))



import typing as t
from tools.toolbox.db import Db 
from typing import Any

db = Db()

def get_variable(name:str) -> Any:
    response = db.get_by_id_or_name(name, "variables")
    return response
    
def get_all_variables() -> Any:
    response = db.get_data("variables")
    return response

def create_variable(data:dict) -> Any:
    response = db.post_data(data, "variables")
    return response

def update_variable(name:str, new_value:t.Union[str,int,bool]) -> Any:
    variable = {"value": new_value}
    response = db.update_data(name, variable, "variables")
    return response

def delete_variable(name:str) -> Any:
    response = db.delete_data(name, "variables")
    return response



"""
Galago Tools - Open Source Lab Automation GRPC Library
"""

from .version import __version__, VERSION_FULL, VERSION_SHORT, VERSION_MAJOR, VERSION_MINOR, VERSION_PATCH

# Export version information to make it available at the package level
__all__ = ['__version__', 'VERSION_FULL', 'VERSION_SHORT', 'VERSION_MAJOR', 'VERSION_MINOR', 'VERSION_PATCH']


import logging
import os
import tempfile

from tools.base_server import ToolServer, serve
from tools.grpc_interfaces.opentrons2_pb2 import Command, Config
from google.protobuf import json_format
from tools.app_config import Config as AppConfig 
from .driver import Ot2Driver
import argparse
import json 

class Opentrons2Server(ToolServer):
    toolType = "opentrons2"
    driver: Ot2Driver
    config: Config
    
    def __init__(self) -> None:
        super().__init__()
        self.app_config = AppConfig()
        self.driver_config: Config 
        self.process_images: bool = True 
          
    def _configure(self, config: Config) -> None:
        self.driver_config = config
        # self.driver = Ot2Driver(robot_ip=config.robot_ip, robot_port=config.robot_port)
        # self.driver.ping()

    def _create_executable_script(self, script_content: str, variables: dict) -> str:
        """
        Create an executable script by injecting variables at the top of the script.
        """
        # Create the variables definition section
        variables_section = "# Injected variables\n"
        
        for key, var_data in variables.items():
            # Extract the actual variable name and value from the variable data structure
            if isinstance(var_data, dict) and 'name' in var_data and 'value' in var_data:
                var_name = var_data['name']
                raw_value = var_data['value']
                var_type = var_data.get('type', 'string')
                
                # Parse the value based on its type
                if var_type == 'array':
                    # Parse JSON array string
                    try:
                        parsed_value = json.loads(raw_value)
                        variables_section += f'{var_name} = {repr(parsed_value)}\n'
                    except json.JSONDecodeError:
                        logging.warning(f"Failed to parse array variable {var_name}: {raw_value}")
                        variables_section += f'{var_name} = []\n'
                        
                elif var_type == 'boolean':
                    # Parse boolean string
                    bool_value = raw_value.lower() in ('true', '1', 'yes', 'on')
                    variables_section += f'{var_name} = {bool_value}\n'
                    
                elif var_type == 'number':
                    # Parse number string
                    try:
                        if '.' in str(raw_value):
                            parsed_value = float(raw_value)
                        else:
                            parsed_value = int(raw_value)
                        variables_section += f'{var_name} = {parsed_value}\n'
                    except (ValueError, TypeError):
                        logging.warning(f"Failed to parse number variable {var_name}: {raw_value}")
                        variables_section += f'{var_name} = 0\n'
                        
                elif var_type == 'string':
                    variables_section += f'{var_name} = "{raw_value}"\n'
                    
                else:
                    # Fallback for unknown types
                    variables_section += f'{var_name} = "{raw_value}"\n'
                    
            else:
                # Handle case where key is the variable name and var_data is the direct value
                var_name = str(key)
                if isinstance(var_data, str):
                    variables_section += f'{var_name} = "{var_data}"\n'
                elif isinstance(var_data, (list, dict)):
                    variables_section += f'{var_name} = {repr(var_data)}\n'
                else:
                    variables_section += f'{var_name} = {var_data}\n'
        
        variables_section += "\n# End injected variables\n\n"
        
        # Combine variables with the script content
        processed_script = variables_section + script_content
        
        # Create temporary file
        with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".py") as f:
            f.write(processed_script)
            logging.info(f"Created executable script: {f.name}")
            return f.name

    def RunProgram(self, params: Command.RunProgram) -> None:
        """
        Execute a Python script with variables passed directly in the request.
        """
        script_content = params.script_content
        variables_dict = json_format.MessageToDict(params.variables) if params.variables else {}
        logging.info(f"Running program with {len(variables_dict)} variables")
        try:
            # Process the script and create executable file with injected variables
            executable_script = self._create_executable_script(script_content, variables_dict)
            
            # Execute the script on the OT-2
            self.driver.start_protocol(protocol_file=executable_script)
            
            # Cleanup temporary file
            try:
                os.unlink(executable_script)
                logging.info(f"Cleaned up temporary script: {executable_script}")
            except OSError as e:
                logging.warning(f"Could not delete temporary file {executable_script}: {e}")
                
        except Exception as e:
            logging.error(f"Error running program: {e}")
            raise

    def Pause(self, params: Command.Pause) -> None:
        logging.info("Pausing program")
        self.driver.pause_protocol()

    def Resume(self, params: Command.Resume) -> None:
        logging.info("Resuming program")
        self.driver.resume_protocol()

    def Cancel(self, params: Command.Cancel) -> None:
        logging.info("Canceling program")
        self.driver.cancel_protocol()

    def ToggleLight(self, params: Command.ToggleLight) -> None:
        logging.info("Toggling light")
        self.driver.toggle_light()

    def EstimateRunProgram(self, params: Command.RunProgram) -> int:
        return 1

    def EstimatePause(self, params: Command.Pause) -> int:
        return 1

    def EstimateResume(self, params: Command.Resume) -> int:
        return 1

    def EstimateCancel(self, params: Command.Cancel) -> int:
        return 1

    def EstimateToggleLight(self, params: Command.ToggleLight) -> int:
        return 1


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument('--port')
    args = parser.parse_args()
    if not args.port:
         raise RuntimeWarning("Port must be provided...")
    serve(Opentrons2Server(), str(args.port))


from datetime import datetime
import os
import logging
from enum import Enum 
from typing import Optional, Any
import sys 
import socket 

class LogType(Enum):
    ERROR = "ERROR",
    WARNING = "WARNING",
    DEBUG = "DEBUG",
    INFO = "INFO",
    PLATE_MOVE = "PLATE_MOVE",
    RUN_START = "RUN_START",
    RUN_END = "RUN_END",
    PLATE_READ = "PLATE_READ",

def get_local_ip() -> Any:
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8", 80))
        local_ip = s.getsockname()[0]
    except Exception:
        local_ip = "127.0.0.1" # Default local IP
    finally:
        s.close()
    return local_ip

def get_shell_command(tool_type:str, port:int) -> list:
    return [sys.executable, '-m', f'tools.{tool_type}.server', f'--port={port}']

def write_trace_log(log_path:Optional[str], log_type:LogType, tool:str,value:str) -> None:

    if not log_path:
        logging.warning("Log folder not configured")
    if log_path is None:
        return
    if not os.path.exists(log_path):
        try:
            os.makedirs(log_path)
        except Exception as e:
            logging.warning(f"Failed to create log folder. {e}")
            return
    file_folder= os.path.join(log_path, datetime.today().strftime('%Y-%m-%d'))
    if(os.path.exists(file_folder) is False):
        logging.debug("folder does not exist. creating folder")
        os.mkdir(file_folder)

    trace_file = os.path.join(file_folder, "trace_log.txt")
    error_file = os.path.join(file_folder, "error_log.txt")

    try:
        if os.path.exists(trace_file) is False:
            with open(trace_file, 'w+') as f:
                f.write('Time,Tool,Value\n')
        if(log_type == LogType.ERROR):
            if os.path.exists(error_file) is False:
                with open(error_file, 'w+') as f:
                    f.write('Time,Tool,Error\n')
    except Exception as e:
        logging.debug(e)
        return
    try:
        with open(trace_file, 'a') as f:
            f.write(str(datetime.today())+","+str(log_type.name)+","+tool+","+value+"\n")
        if log_type == LogType.ERROR:
            with open(error_file, 'a') as f:
                 f.write(str(datetime.today())+","+str(log_type.name)+","+tool+","+value+"\n")
    except Exception as e:
        logging.debug(e)
        return

def list_available_tools() -> list:
    import os
    import importlib
    import inspect
    from pkgutil import iter_modules
    
    tool_list = []
    tool_path = os.path.join(os.path.dirname(__file__))
    
    for module_info in iter_modules([tool_path]):
        module_name = module_info.name
        server_module_path = f"tools.{module_name}.server"
        
        try:
            # Try to import the server module from the tool directory
            server_module = importlib.import_module(server_module_path)
            
            # Check if any class in the server module inherits from ToolServer
            has_toolserver_subclass = False
            for name, obj in inspect.getmembers(server_module, inspect.isclass):
                # Check if the class is defined in this server module (not imported)
                if obj.__module__ == server_module_path:
                    # Get the method resolution order to check inheritance
                    for base_class in inspect.getmro(obj):
                        if base_class.__name__ == 'ToolServer' and base_class is not obj:
                            has_toolserver_subclass = True
                            break
                    if has_toolserver_subclass:
                        break
            
            if has_toolserver_subclass:
                tool_list.append(module_name)
                
        except ImportError:
            continue
        except Exception:
            continue 
    return tool_list


def get_tool_server_info(tool_name: str) -> dict:
    """
    Get detailed information about a specific tool server and its commands.
    
    Args:
        tool_name: The name of the tool (e.g., 'opentrons2', 'plateloc')
    
    Returns:
        Dictionary containing tool server information including commands and their parameters
    """
    import importlib
    import inspect
    from google.protobuf.message import Message
    
    try:
        # Import the server module
        server_module_path = f"tools.{tool_name}.server"
        server_module = importlib.import_module(server_module_path)
        
        # Import the corresponding protobuf module for Command types
        try:
            pb_module_path = f"tools.grpc_interfaces.{tool_name}_pb2"
            pb_module = importlib.import_module(pb_module_path)
            command_class = getattr(pb_module, 'Command', None)
        except ImportError:
            pb_module = None
            command_class = None
        
        # Find the ToolServer subclass
        tool_server_class = None
        for name, obj in inspect.getmembers(server_module, inspect.isclass):
            if obj.__module__ == server_module_path:
                for base_class in inspect.getmro(obj):
                    if base_class.__name__ == 'ToolServer' and base_class is not obj:
                        tool_server_class = obj
                        break
                if tool_server_class:
                    break
        
        if not tool_server_class:
            return {"error": f"No ToolServer subclass found in {server_module_path}"}
        
        # Get basic tool information
        tool_info = {
            "tool_name": tool_name,
            "class_name": tool_server_class.__name__,
            "module_path": server_module_path,
            "tool_type": getattr(tool_server_class, 'toolType', tool_name),
            "commands": {},
            "estimate_methods": {},
            "other_methods": {}
        }
        
        # Helper function to get protobuf message fields
        def get_protobuf_fields(message_class : type) -> list[str]:
            """Extract field names from a protobuf message class."""
            if not message_class or not issubclass(message_class, Message):
                return []
            
            # Method 1: Use DESCRIPTOR if available
            if hasattr(message_class, 'DESCRIPTOR'):
                try:
                    descriptor = message_class.DESCRIPTOR
                    return [field.name for field in descriptor.fields]
                except Exception:
                    pass
            
            # Method 2: Look for field number constants
            field_names = []
            for attr_name in dir(message_class):
                if attr_name.endswith('_FIELD_NUMBER'):
                    # Convert SCRIPT_CONTENT_FIELD_NUMBER -> script_content
                    field_name = attr_name[:-13].lower()  # Remove _FIELD_NUMBER
                    field_names.append(field_name)
            
            return field_names
        
        # Analyze all methods in the class
        for method_name, method_obj in inspect.getmembers(tool_server_class):
            if (method_name.startswith('_') or 
                not callable(method_obj)):
                continue  # Skip private methods and non-callables
                
            # Get method signature
            try:
                sig = inspect.signature(method_obj)
                params = list(sig.parameters.values())[1:]  # Skip 'self' parameter
            except (ValueError, TypeError):
                continue
            
            # Check if this is a command method (has params with Command type)
            command_param = None
            for param in params:
                if param.name == 'params' and param.annotation != inspect.Parameter.empty:
                    annotation = param.annotation
                    annotation_str = str(annotation)
                    
                    # Check if this is a protobuf Message subclass
                    is_command_type = False
                    try:
                        # Check if annotation is a class and subclass of Message
                        if (inspect.isclass(annotation) and 
                            issubclass(annotation, Message)):
                            is_command_type = True
                    except (TypeError, AttributeError):
                        # Fallback: check string representation
                        is_command_type = (
                            hasattr(annotation, '__module__') and 
                            annotation.__module__ and 
                            '_pb2' in annotation.__module__ and
                            'Command.' in annotation_str
                        )
                    
                    if is_command_type:
                        # Extract the command type name
                        if hasattr(annotation, '__qualname__'):
                            type_name = annotation.__qualname__
                        else:
                            type_name = annotation_str.split('.')[-1] if '.' in annotation_str else str(annotation)
                        
                        # Get fields using protobuf-specific methods
                        fields = get_protobuf_fields(annotation)
                        
                        command_param = {
                            "name": param.name,
                            "type": type_name,
                            "annotation": annotation_str,
                            "fields": fields
                        }
                        
                        break
            
            if command_param:
                tool_info["commands"][method_name] = {
                    "method_name": method_name,
                    "parameter": command_param,
                    "docstring": inspect.getdoc(method_obj),
                    "signature": str(sig)
                }
            elif method_name.startswith('Estimate'):
                # Estimate methods
                tool_info["estimate_methods"][method_name] = {
                    "method_name": method_name,
                    "return_type": str(sig.return_annotation) if sig.return_annotation != inspect.Signature.empty else "Unknown",
                    "signature": str(sig),
                    "docstring": inspect.getdoc(method_obj)
                }
            else:
                # Other public methods (skip some common inherited methods)
                if method_name not in ['__init__', '__class__', '__dict__', '__doc__', '__module__', '__weakref__']:
                    tool_info["other_methods"][method_name] = {
                        "method_name": method_name,
                        "signature": str(sig),
                        "docstring": inspect.getdoc(method_obj)
                    }
        
        # Get Command class information if available
        if command_class:
            tool_info["command_class_info"] = {
                "class_name": command_class.__name__,
                "available_commands": []
            }
            
            # Get all nested Command classes
            for attr_name in dir(command_class):
                if attr_name.startswith('_'):
                    continue
                    
                attr = getattr(command_class, attr_name)
                if inspect.isclass(attr) and issubclass(attr, Message):
                    fields = get_protobuf_fields(attr)
                    command_info = {
                        "name": attr_name,
                        "full_name": f"Command.{attr_name}",
                        "fields": fields
                    }
                    tool_info["command_class_info"]["available_commands"].append(command_info)
        
        return tool_info
        
    except ImportError as e:
        return {"error": f"Could not import tool '{tool_name}': {e}"}
    except Exception as e:
        return {"error": f"Error analyzing tool '{tool_name}': {e}"}

def print_tool_server_info(tool_name: str) -> None:
    """
    Print formatted information about a tool server.
    """
    info = get_tool_server_info(tool_name)
    
    if "error" in info:
        print(f"Error: {info['error']}")
        return
    
    print(f"\n=== Tool Server Information: {info['tool_name']} ===")
    print(f"Class Name: {info['class_name']}")
    print(f"Module Path: {info['module_path']}")
    print(f"Tool Type: {info['tool_type']}")
    
    if info.get('command_class_info'):
        cmd_class = info['command_class_info']
        print(f"\n--- Available Protobuf Commands ({len(cmd_class['available_commands'])}) ---")
        for cmd in cmd_class['available_commands']:
            fields_str = f" (fields: {', '.join(cmd['fields'])})" if cmd['fields'] else " (no fields)"
            print(f"• {cmd['name']}{fields_str}")
    
    print("\n" + "="*50)


if __name__ == "__main__":
    # print("Available tools:", list_available_tools())
    print_tool_server_info("pf400")


import typing as t
import os
import datetime
import logging
import time
import traceback
import shutil
import xml.etree.ElementTree as ET
import queue
import threading
from tools.base_server import ABCToolDriver
from typing import Union, Optional
try:
    import pythoncom
    import win32com.client as win32
except Exception:
    # The driver will error if there's no pythoncom
    pass

DEFAULT_EXPERIMENT_DIR = "C:\\cytation_output"
DEFAULT_PROTOCOL_DIR = "C:\\cytation\\protocols"
CYTATION5_READER_TYPE = 21 #Default to Cytation 5


class CytationDriver(ABCToolDriver):
    def __init__(
        self,
        protocol_dir: str,
        experiment_dir: str,
        reader_type: Union[str,int] = CYTATION5_READER_TYPE,
    ) -> None:
        self.experiment_dir = experiment_dir
        self.protocol_dir = protocol_dir
        self.reader_type = reader_type
        self._command_lock: threading.Lock = threading.Lock()
        self.command_queue: queue.Queue = queue.Queue()
        self.command_response_queue: queue.Queue = queue.Queue()
        self.live: bool = False
        self.live_message: str = ""
        self.execution_thread: Optional[threading.Thread] = None
        self.start()

    def start(self) -> None:
        self.kill_processes("Gen5.exe")
        self.live = True
        self.live_message = ""
        self.execution_thread = threading.Thread(target=self.execute_cytation_commands)
        self.execution_thread.daemon = True
        self.execution_thread.start()

    

    def execute_cytation_commands(self) -> None:
        logging.info("Starting cytation command thread")

        try:
            # Required for running COM objects in threads.
            # https://mail.python.org/pipermail/python-win32/2008-June/007788.html
            pythoncom.CoInitialize()
            app = win32.Dispatch("Gen5.Application")
            app.ConfigureUSBReader(self.reader_type, "")
            while self.live:
                while not self.command_queue.empty() > 0:
                    with self._command_lock:
                        command_obj = self.command_queue.get()
                    response = self.execute_command(
                        app,
                        command_obj["command"],
                        command_obj["params"],
                    )
                    with self._command_lock:
                        self.command_response_queue.put(
                            {"command": command_obj["command"], "response": response}
                        )

                time.sleep(0.25)

        except Exception as e:
            logging.warning("Cytation command thread has errored")
            logging.error(traceback.format_exc())
            self.live = False
            self.live_message = str(e)
            pythoncom.CoUninitialize()

    def schedule_command(self, command: str, params: dict[str, t.Any] = {}) -> None:
        logging.info(f"Scheduling command {command}, {params}")
        with self._command_lock:
            self.command_queue.put({"command": command, "params": params})

    def save_picture_builders(self, plate: t.Any, experiment_name: str) -> None:
        logging.info("Saving picture builders...")
        # picture output (do we need this if we get raw tiff files?)
        picture_builder_names = win32.VARIANT(
            pythoncom.VT_VARIANT | pythoncom.VT_BYREF, []
        )
        plate.GetPictureExportNames(False, picture_builder_names)
        logging.info(f"Picture builder names: {picture_builder_names.value}")
        for picture_builder in picture_builder_names.value:
            logging.info(f"Picture builder name: {picture_builder}")
            # write to location defined in protcol
            xml_response = plate.PictureExport(picture_builder)
            # get location and move if not in experiment folder
            save_dir = ET.fromstring(xml_response).findall("Folder")[0].text
            image_name = ET.fromstring(xml_response).findall("Image")[0].text
            if save_dir:
                new_dir = f"{self.experiment_dir}\\{experiment_name}\\{image_name}"
                # check if dir exists
                if not os.path.exists(new_dir):
                    os.makedirs(new_dir)
                # move all files in save_dir that have experiment_name in their name to new_dir
                filenames = ET.fromstring(xml_response).findall("PictureFile")
                for file_element in filenames:
                    filename = file_element.text
                    shutil.move(f"{save_dir}\\{filename}", f"{new_dir}\\{filename}")
                # shutil.move(save_dir, new_dir)

    def save_export_builders(self, plate: t.Any, experiment_name: str) -> None:
        logging.info("Saving export builders...")
        # output exports (metadata and analysis results)
        export_builder_names = win32.VARIANT(
            pythoncom.VT_VARIANT | pythoncom.VT_BYREF, []
        )
        plate.GetFileExportNames(False, export_builder_names)
        logging.info(f"Export builder names: {export_builder_names.value}")
        # export data to location of our choosing
        for export_builder in export_builder_names.value:
            logging.info(f"Export builder name: {export_builder}")
            plate.FileExportEx(
                export_builder,
                f"{self.experiment_dir}\\{experiment_name}\\{experiment_name}_{export_builder}.csv",
            )

    def move_raw_data(self, plate: t.Any, experiment_name: str) -> None:
        logging.info("Raw data export...")
        raw_image_output_folders = win32.VARIANT(
            pythoncom.VT_VARIANT | pythoncom.VT_BYREF, []
        )
        plate.GetImageFolderPaths(raw_image_output_folders)
        image_exp_dirname = os.path.dirname(
            os.path.dirname(raw_image_output_folders.value[0])
        )
        logging.info("Protocol image export folder: " + image_exp_dirname)
        if image_exp_dirname != f"{self.experiment_dir}\\{experiment_name}":
            logging.info("Moving raw data to main experiment output folder...")
            folder_name = os.path.basename(
                os.path.dirname(raw_image_output_folders.value[0])
            )
            shutil.move(
                os.path.dirname(raw_image_output_folders.value[0]),
                f"{self.experiment_dir}\\{experiment_name}\\{folder_name}",
            )

    def execute_command(
        self, app: t.Any, command: str, command_params: dict[str, t.Any]
    ) -> Optional[int]:
        logging.info(f"Executing command {command}")

        if command == "start_read":
            protocol_file = command_params["protocol_file"]
            experiment_name = command_params["experiment_name"]
            well_addresses = command_params["well_addresses"]

            logging.info(f"Wells To Image: {well_addresses}")

            date = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
            # create experiment name if blank
            if experiment_name == "":
                experiment_name = f"auto_{protocol_file}_{date}"
            # check if directory exists
            # if os.path.exists(os.path.join(self.experiment_dir, experiment_name)):
            #     experiment_name = f"{experiment_name}_{date}"
    
            protocol_path = f"{self.protocol_dir}\\{protocol_file}"
            logging.info(f"start_read called with file {protocol_path}")
            experiment_path = (
                f"{self.experiment_dir}\\{experiment_name}\\{experiment_name}.xpt"
            )
            if len(experiment_path) > 250:
                experiment_path= f"{experiment_path[:240]}.xpt"
            logging.info(f"Saving new experiment file to {experiment_path}")
            experiment = app.NewExperiment(protocol_path)
            experiment.SaveAs(experiment_path)
            plate = experiment.plates.GetPlate(1)

            if well_addresses is not None and len(well_addresses) > 0:
                logging.info(f"Setting wells to image to {well_addresses}")
                xml_string = '<BTIPartialPlate Version="1.00"><SingleBlock>No</SingleBlock><Wells>'
                for well_addy in well_addresses:
                    xml_string += f"<Well>{well_addy}</Well>"
                xml_string += "</Wells></BTIPartialPlate>"
                logging.info(f"XML string: {xml_string}")
                plate.SetPartialPlate(xml_string)

            logging.info("Starting read...")
            # We've seen this function call hang indefinitely, which is bad.
            # It's not easy to kill threads in Python, but we might want to look into it,
            # in case the execution thread hangs.
            monitor = plate.StartRead

            # Wait for a moment so the ReadInProgress flag has time to set itself.
            time.sleep(5)

            times = 0
            while monitor.ReadInProgress and self.live:
                if times % 60 == 0:
                    logging.info("Cytation read in progress...")
                times += 1
                time.sleep(1)

            logging.info("Cytation read complete...")

            # output picture builders (PNGs usually)
            self.save_picture_builders(plate, experiment_name)

            # output csvs with metadata and/or analysis results
            self.save_export_builders(plate, experiment_name)

            # move raw data to main experiment output folder if necessary
            # FOR NOW WE ARE NOT DOING THIS, WANT TO ADD IN MAKING IT OPTIONAL
            # self.move_raw_data(plate, experiment_name)

        elif command == "open_carrier":
            app.CarrierOut()
        elif command == "close_carrier":
            app.CarrierIn()
        elif command == "test_reader_communication":
            reader_state: int = app.TestReaderCommunication
            logging.info(f"Got reader state {reader_state}")
            return reader_state
        return None

    def wait_for_command(
        self, command: str, timeout:Optional[int] = None
    ) -> Optional[Union[str,int]]:
        times = 0
        start_time = time.time()

        while self.live:
            seconds_spent_waiting = int(time.time() - start_time)
            if timeout and seconds_spent_waiting > timeout:
                raise Exception(
                    f"Command {command} has timed out. Please reset the driver."
                )

            if not self.command_response_queue.empty() > 0:
                with self._command_lock:
                    command_obj: dict[str, str] = self.command_response_queue.get()
                    if command_obj["command"] == command:
                        logging.info(
                            f"Command {command} has completed. Waited for {seconds_spent_waiting} seconds"
                        )
                        return command_obj["response"]
                    else:
                        logging.warning(
                            f"Unexpected command {command_obj['command']} received"
                        )
            times += 1
            if not self.live:
                break

            if times % 60 == 0:
                logging.info(
                    f"Waiting for Cytation {command} command...({seconds_spent_waiting}s)"
                )
            time.sleep(1)

        if not self.live:
            raise Exception(
                f"Cytation driver has crashed. Please reset the driver. {self.live_message}"
            )
        return None

    def start_read(
        self,
        protocol_file: str,
        experiment_name: str = "",
        well_addresses: t.Optional[t.List[str]] = None,
    ) -> None:
        if not protocol_file.endswith(".prt"):
            protocol_file = f"{protocol_file}.prt"
        self.schedule_command(
            "start_read",
            {
                "protocol_file": protocol_file,
                "experiment_name": experiment_name,
                "well_addresses": well_addresses,
            },
        )

        self.wait_for_command("start_read")

    def open_carrier(self) -> None:
        self.schedule_command(
            "open_carrier",
        )
        self.wait_for_command("open_carrier")

    def close_carrier(self) -> None:
        self.schedule_command("close_carrier")
        self.wait_for_command("close_carrier")

    def close(self) -> None:
        self.live = False
        if self.execution_thread:
            self.execution_thread.join()

    """
        Test whether we can connect to the Cytation reader.
    """

    def _test_reader_communication(self) -> Optional[Union[str, int]]:
        self.schedule_command(
            "test_reader_communication",
        )
        return self.wait_for_command("test_reader_communication", timeout=8)

    def verify_reader_communication(self) -> None:
        reader_state: Optional[Union[str,int]] = self._test_reader_communication()

        if reader_state != 1:
            raise Exception(f"Expected reader state 1. Got {reader_state}")

    def __del__(self) -> None:
        """Disconnect from the device when the driver is destroyed."""
        self.close()



import os

class Data():
    def __init__(self)-> None:
        return
    
    # @staticmethod
    # def get_media_exchange_logs_by_date(date:str) -> list[dict]:
    #     conf = Config()
    #     conf.load_app_config()
    #     data_dir = conf.app_config.data_folder
    #     if not data_dir:
    #         raise RuntimeError("Data directory can't be null")
    #     file= os.path.join(data_dir,"media_addition",f"media_addition_{date}.txt")
    #     result = []
    #     if os.path.exists(file):
    #         with open(file, mode='r') as f:
    #             next(f)
    #             lines = f.readlines()
    #             for line in lines:
    #                 print(f"Line is {line}")
    #                 row : dict = {}
    #                 line_splitted = line.split(",")
    #                 row = {
    #                     "source_barcode":line_splitted[0],
    #                     "destination_name":line_splitted[1],
    #                     "destination_barcode":line_splitted[2],
    #                     "source_wells":line_splitted[3],
    #                     "percent_exchange":line_splitted[4],
    #                     "new_tips":line_splitted[5],
    #                     "created_on":line_splitted[6]
    #                 }
    #                 result.append(row)
                
    #             return list(reversed(result))
    #     else:
    #         return list({})
    #For now we are assuming there is only 1 liconic per workcell, we should id them in case there are more.
    # @staticmethod
    # def get_liconic_sensor_data(liconic_id:str, date:str) -> dict[str, t.Any]:
    #     conf = Config()
    #     conf.load_app_config()
    #     if not conf.app_config.data_folder:
    #         return {'times':[],'co2_values':[]}
    #     liconic_folder = join(conf.app_config.data_folder,"sensors","liconic",date)
    #     response = {}
    #     time_array : list[str]= []
    #     co2_array : list[str] = []
    #     if os.path.exists(liconic_folder) is False:
    #         return {'times':[],'co2_values':[]}
    #     else:
    #         liconic_file = join(liconic_folder,f"{liconic_id}_co2.txt")
    #         with open(liconic_file, mode='r') as file:
    #             next(file)
    #             lines = file.readlines()
    #             for line in lines:
    #                 line_array = line.split(',')
    #                 time = line_array[0]
    #                 co2 = line_array[1].replace("\n","")
    #                 time_array.append(time)
    #                 co2_array.append(co2)
    #             response["times"]=time_array
    #             response["co2_values"]=co2_array
    #         return response
        
    # @staticmethod
    # def get_ot2_images_by_date(date:str) -> list[str]:
    #     conf = Config()
    #     conf.load_app_config()
    #     if not conf.app_config.data_folder:
    #         return []
    #     ot2_images_folder = join(conf.app_config.data_folder,"images","ot2",date)
    #     images = []
    #     if os.path.exists(ot2_images_folder):
    #         images = [x for x in os.listdir(ot2_images_folder) if x.endswith(".jpg") and "predicted" not in x]
    #         # Sort by creation time
    #         images.sort()
    #         #images.sort(key=lambda x: os.path.getctime(os.path.join(ot2_images_folder, x)))

    #     return images
    
    # @staticmethod
    # def get_ot2_image_bites_by_file_name(date:str, file_name:str) -> dict:
    #     conf = Config()
    #     conf.load_app_config()
    #     if not conf.app_config.data_folder:
    #         return {"created_on":"null","img_bytes":"null"}
    #     ot2_image = join(conf.app_config.data_folder,"images","ot2",date,file_name)
    #     if os.path.exists(ot2_image):
    #         try:
    #             time_pattern = r"\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}"
    #             match = re.search(time_pattern, file_name)
    #             if match:
    #                 datetime_string = match.group()
    #                 dt = datetime.strptime(datetime_string, "%Y-%m-%d_%H-%M-%S")
    #                 formatted_datetime = dt.strftime("%m-%d-%Y %H:%M:%S")
    #             else:
    #                 modified_time = os.path.getmtime(ot2_image)
    #                 formatted_datetime = datetime.fromtimestamp(modified_time).strftime('%m-%d-%Y %H:%M:%S')
    #             with open(ot2_image, 'rb') as image_file:
    #                 image_bytes = image_file.read()
    #                 econcoded_string = base64.b64encode(image_bytes).decode('utf-8')
    #             return  {"created_on":formatted_datetime,"img_bytes":f"data:image/jpeg;base64,{econcoded_string}"}
                
    #         except (OSError, IOError) as e:
    #             print(f"Error reading image file {ot2_image}: {e}")
    #             return {"created_on":"null","img_bytes":"null"}
    #         except Exception as e:
    #             print(f"Unexpected error: {e}")
    #             return {"created_on":"null","img_bytes":"null"}
    #     else:
    #         return  {"created_on":"null","img_bytes":"null"}

    # @staticmethod
    # def get_ot2_processed_image_bites_by_file_name(date:str, file_name:str) -> dict:
    #     conf = Config()
    #     conf.load_app_config()
    #     if not conf.app_config.data_folder:
    #         return {"created_on":"null","img_bytes":"null"}
    #     ot2_image = join(conf.app_config.data_folder,"images","ot2",date,file_name)
    #     if not os.path.exists(ot2_image):
    #         return  {"created_on":"null","img_bytes":"null"}
    #     try:
    #         modified_time = os.path.getmtime(ot2_image)
    #         formatted_time = datetime.fromtimestamp(modified_time).strftime('%m-%d-%Y %H:%M:%S')
    #         with open(ot2_image, 'rb') as image_file:
    #             image_bytes = image_file.read()
    #             econcoded_string = base64.b64encode(image_bytes).decode('utf-8')
    #         return  {"created_on":formatted_time,"img_bytes":f"data:image/jpeg;base64,{econcoded_string}"}
    #     except (OSError, IOError) as e:
    #         logging.warning(f"Error reading image file {ot2_image}: {e}")
    #         return  {"created_on":"null","img_bytes":"null"}
    #     except Exception as e:
    #         logging.warning(f"Unexpected error: {e}")
    #         return  {"created_on":"null","img_bytes":"null"}
    
    @staticmethod
    def validate_folder(folder_path:str) -> bool:
        if os.path.exists(folder_path):
            return True
        else:
            return False
        



import logging
from tools.base_server import ABCToolDriver
import requests
import time
from typing import Optional 
from PIL import Image
import io
import os 

class Ot2Driver(ABCToolDriver):
    def __init__(self, robot_ip: str, robot_port: int = 31950) -> None:
        self.robot_ip: str = robot_ip
        self.robot_port: int = robot_port

        self.run_id: Optional[str] = None

        self.base_url = f"http://{robot_ip}:{robot_port}"
        self.headers = {"Opentrons-Version": "2"}

    def ping(self) -> None:
        """Test connection to the OT-2 robot."""
        response = requests.get(
            url=f"{self.base_url}/health",
            headers=self.headers,
        )

        if response.status_code != 200:
            raise Exception(f"Ping failed with status {response.status_code}: {response.text}")
        else:
            logging.info("OT-2 ping successful")
    
    def toggle_light(self) -> None:
        """Toggle the OT-2 deck lights on/off."""
        try:
            # Get current light status
            light_status_response = requests.get(
                url=f"{self.base_url}/robot/lights",
                headers=self.headers,
            )
            if not light_status_response.ok:
                raise Exception(f"Failed to get light status: {light_status_response.text}")
            
            light_status: bool = light_status_response.json()['on']
            logging.info(f"Current light status: {'ON' if light_status else 'OFF'}")

            # Toggle light status
            toggle_light_response = requests.post(
                url=f"{self.base_url}/robot/lights",
                headers=self.headers,
                json={"on": not light_status}
            )
            if not toggle_light_response.ok:
                raise Exception(f"Failed to toggle lights: {toggle_light_response.text}")
            
            new_status = "ON" if not light_status else "OFF"
            logging.info(f"Lights toggled to: {new_status}")
            
        except Exception as e:
            logging.error(f"Error toggling lights: {e}")
            raise

    def upload_and_schedule_protocol(self, protocol_file: str) -> str:
        """Upload protocol file and create a scheduled run."""
        try:
            # Upload protocol
            with open(protocol_file, 'rb') as f:
                upload_response = requests.post(
                    url=f"{self.base_url}/protocols",
                    files={
                        "files": ('protocol.py', f, "text/x-python-script"),
                    },
                    headers=self.headers,
                )
            
            if not upload_response.ok:
                raise Exception(f"Protocol upload failed: {upload_response.text}")
                
            try:
                protocol_id = upload_response.json()['data']['id']
                logging.info(f"Protocol uploaded with ID: {protocol_id}")
            except KeyError:
                raise Exception(f"Invalid upload response format: {upload_response.text}")

            # Create a run for the protocol
            create_run_response = requests.post(
                url=f"{self.base_url}/runs",
                json={
                    "data": {
                        "protocolId": protocol_id
                    }
                },
                headers=self.headers,
            )
            
            if not create_run_response.ok:
                raise Exception(f"Failed to create run: {create_run_response.text}")
                
            run_id: str = create_run_response.json()['data']['id']
            self.run_id = run_id
            logging.info(f'Created run with ID: {self.run_id}')
            
            return run_id
            
        except Exception as e:
            logging.error(f"Error uploading and scheduling protocol: {e}")
            raise

    def start_run(self, run_id: str) -> None:
        """Start a scheduled protocol run."""
        try:
            start_response = requests.post(
                url=f"{self.base_url}/runs/{run_id}/actions",
                headers=self.headers,
                json={"data": {"actionType": "play"}}
            )
            if not start_response.ok:
                raise Exception(f"Failed to start run: {start_response.text}")
            
            logging.info(f"Started run: {run_id}")
            
        except Exception as e:
            logging.error(f"Error starting run {run_id}: {e}")
            raise

    def pause_protocol(self) -> None:
        """Pause the currently running protocol."""
        if not self.run_id:
            logging.warning("No active run to pause")
            return
            
        try:
            pause_response = requests.post(
                url=f"{self.base_url}/runs/{self.run_id}/actions",
                headers=self.headers,
                json={"data": {"actionType": "pause"}}
            )
            if not pause_response.ok:
                raise Exception(f"Failed to pause run: {pause_response.text}")
            
            logging.info(f"Paused run: {self.run_id}")
            
        except Exception as e:
            logging.error(f"Error pausing run: {e}")
            raise
    
    def resume_protocol(self) -> None:
        """Resume a paused protocol."""
        if not self.run_id:
            logging.warning("No active run to resume")
            return
            
        try:
            resume_response = requests.post(
                url=f"{self.base_url}/runs/{self.run_id}/actions",
                headers=self.headers,
                json={"data": {"actionType": "play"}}
            )
            if not resume_response.ok:
                raise Exception(f"Failed to resume run: {resume_response.text}")
            
            logging.info(f"Resumed run: {self.run_id}")
            
        except Exception as e:
            logging.error(f"Error resuming run: {e}")
            raise
    
    def cancel_protocol(self) -> None:
        """Cancel and delete the currently running protocol."""
        if not self.run_id:
            logging.warning("No active run to cancel")
            return
            
        try:
            # Stop the run
            stop_response = requests.post(
                url=f"{self.base_url}/runs/{self.run_id}/actions",
                headers=self.headers,
                json={"data": {"actionType": "stop"}}
            )
            if not stop_response.ok:
                raise Exception(f"Failed to stop run: {stop_response.text}")
            
            # Delete the run
            delete_response = requests.delete(
                url=f"{self.base_url}/runs/{self.run_id}",
                headers=self.headers,
            )
            if not delete_response.ok:
                logging.warning(f"Failed to delete run: {delete_response.text}")
                # Don't raise exception here as the run is already stopped
            
            logging.info(f"Cancelled run: {self.run_id}")
            self.run_id = None
            
        except Exception as e:
            logging.error(f"Error cancelling run: {e}")
            raise
    
    def get_run_status(self, run_id: str) -> dict:
        """Get the current status of a protocol run."""
        try:
            response = requests.get(
                url=f"{self.base_url}/runs/{run_id}",
                headers=self.headers,
                timeout=30
            )
            if not response.ok:
                raise Exception(f"Failed to get run status: {response.text}")
            
            return response.json()
            
        except Exception as e:
            logging.error(f"Error getting run status for {run_id}: {e}")
            raise

    def wait_for_completion(self, run_id: str, timeout: int = 1800) -> None:
        """Wait for a protocol run to complete."""
        start_time = time.time()
        
        while time.time() - start_time < timeout:
            try:
                run_data = self.get_run_status(run_id)
                status = run_data['data']['status']
                
                logging.info(f"Run {run_id} status: {status}")
                
                if status == 'succeeded':
                    logging.info(f"Run {run_id} completed successfully")
                    return
                elif status in ['stopped', 'failed', 'blocked-by-open-door']:
                    raise Exception(f"Run {run_id} failed with status: {status}")
                
                time.sleep(2)  # Check every 2 seconds
                
            except Exception as e:
                if "failed with status" in str(e):
                    raise  # Re-raise status failures
                logging.warning(f"Error checking run status: {e}")
                time.sleep(5)  # Wait longer on communication errors
        
        raise TimeoutError(f"Run {run_id} did not complete within {timeout} seconds")

    def start_protocol(self, protocol_file: str, wait_for_completion: bool = True) -> None:
        """
        Execute a protocol file on the OT-2.
        
        Args:
            protocol_file: Path to the Python protocol file
            wait_for_completion: Whether to wait for the protocol to finish
        """
        try:
            logging.info(f"Starting protocol: {protocol_file}")
            
            # Upload and schedule the protocol
            run_id = self.upload_and_schedule_protocol(protocol_file)
            
            # Start the run
            self.start_run(run_id)
            
            # Wait for completion if requested
            if wait_for_completion:
                self.wait_for_completion(run_id)
                
            logging.info("Protocol execution completed successfully")
            
        except Exception as e:
            logging.error(f"Error executing protocol: {e}")
            raise
    
    def take_picture(self, name: str, directory: str) -> str:
        """
        Take a picture using the OT-2 camera and save it to the specified location.
        
        Args:
            name: Filename for the image
            directory: Directory to save the image
            
        Returns:
            Full path to the saved image file
        """
        try:
            # Take picture via API
            response = requests.post(
                url=f"{self.base_url}/camera/picture",
                headers=self.headers,
            )
            if not response.ok:
                raise Exception(f"Failed to take picture: {response.text}")
            
            # Process and save the image
            image = Image.open(io.BytesIO(response.content))
            rotated_image = image.rotate(180)  # OT-2 camera is upside down
            
            # Ensure directory exists
            os.makedirs(directory, exist_ok=True)
            
            # Save image
            file_path = os.path.join(directory, name)
            rotated_image.save(file_path, format="JPEG")
            
            logging.info(f"Picture saved: {file_path}")
            return file_path
            
        except Exception as e:
            logging.error(f"Error taking picture: {e}")
            raise

    def close(self) -> None:
        """Clean up resources and cancel any running protocols."""
        try:
            if self.run_id:
                logging.info("Cleaning up active run on close")
                self.cancel_protocol()
        except Exception as e:
            logging.warning(f"Error during cleanup: {e}")


import logging
import os
import queue
import subprocess
import time
from typing import Any, Tuple

from tools.base_server import ABCToolDriver

VWORKS_PROCESS = "VWorks.exe"


def kill_vworks() -> None:
    """force kill vworks"""
    try:
        subprocess.run(
            ["taskkill", "/F", "/IM", VWORKS_PROCESS],
            check=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        time.sleep(0.1)
    except Exception as e:
        logging.warning(f"Failed to kill VWorks on clean up: {e}")


EventType = Tuple[str, Any]
EventQueue = queue.Queue


if os.name == "nt":
    import comtypes.client as cc  # type: ignore
    import pythoncom

    # from comtypes import GUID
    from comtypes.client import GetEvents

    class VWorksEventSink:
        def __init__(self, event_queue: EventQueue) -> None:
            self.event_queue = event_queue

        def InitializationComplete(self, *args: Any) -> int:
            logging.info(f"Initialization complete: {args}")
            self.event_queue.put(("InitializationComplete", args))
            return 0

        def InitializationCompleteWithCode(self, *args: Any) -> int:
            logging.info(f"Initialization complete with code: {args}")
            self.event_queue.put(("InitializationCompleteWithCode", args))
            return 0

        def LogMessage(self, *args: Any) -> int:
            # This fires frequently with many arguments, so just accept all of them
            # Don't put in queue to reduce overhead
            return 0

        # def MessageBoxAction(self, *args):
        #     message = args[2] if len(args) > 2 else "Unknown message"
        #     print(f"Message box action: {message}")
        #     self.event_queue.put(("MessageBoxAction", message))

        #     # Return explicit button choice (1 = OK, 2 = Cancel)
        #     # For compiler errors, choose Cancel (2) which is recommended
        #     if "compiler errors" in message:
        #         return 2  # Cancel
        #     else:
        #         return 1  # OK

        # def UserMessage(self, *args):
        #     message = args[1] if len(args) > 1 else "Unknown message"
        #     print(f"User message: {message}")
        #     self.event_queue.put(("UserMessage", message))
        #     return 0

        def ProtocolComplete(self, *args: Any) -> None:
            protocol = args[1] if len(args) > 1 else "unknown"
            logging.info(f"Protocol completed: {protocol}")
            self.event_queue.put(("ProtocolComplete", protocol))

        def ProtocolAborted(self, *args: Any) -> None:
            protocol = args[1] if len(args) > 1 else "unknown"
            logging.error(f"Protocol aborted: {protocol}")
            self.event_queue.put(("ProtocolAborted", protocol))

        def UnrecoverableError(self, *args: Any) -> None:
            description = args[1] if len(args) > 1 else "unknown error"
            logging.error(f"Unrecoverable error: {description}")
            self.event_queue.put(("UnrecoverableError", description))

        def RecoverableError(self, *args: Any) -> None:
            description = args[3] if len(args) > 3 else "unknown error"
            logging.error(f"Recoverable error: {description}")
            self.event_queue.put(("RecoverableError", description))

            # If we have action parameters
            if len(args) > 4:
                actionToTake = args[4]
                if hasattr(actionToTake, "value"):
                    actionToTake.value = 0  # 0=Abort, 1=Retry, 2=Ignore

            if len(args) > 5:
                vworksHandlesError = args[5]
                if hasattr(vworksHandlesError, "value"):
                    vworksHandlesError.value = True  # Let VWorks handle the error

    # This should match the GUID for _IVWorks4APIEvent in the registry
    # IVWorks4APIEvent = GUID("{EB350F99-1AC0-429E-8213-55D57DC010C1}")  # Example GUID, replace with correct one

else:

    class VWorks4API:  # type: ignore
        def __init__(self) -> None:
            pass

        def Login(self, user: str, password: str) -> None:
            raise NotImplementedError("VWorks4API is not supported on non-Windows platforms.")

        def ShowVWorks(self, show: bool) -> None:
            raise NotImplementedError("VWorks4API is not supported on non-Windows platforms.")

        def Logout(self) -> None:
            raise NotImplementedError("VWorks4API is not supported on non-Windows platforms.")

        def LoadProtocol(self, protocol: str) -> None:
            raise NotImplementedError("VWorks4API is not supported on non-Windows platforms.")

        def RunProtocol(self, protocol: str, times: int) -> None:
            raise NotImplementedError("VWorks4API is not supported on non-Windows platforms.")

        def LoadRunsetFile(self, runset: str) -> None:
            raise NotImplementedError("VWorks4API is not supported on non-Windows platforms.")


class BravoDriver(ABCToolDriver):
    def __init__(self, init_com: bool = False):
        self.live = False
        self.event_queue = queue.Queue()
        self.event_connection = None
        self.driver: VWorks4API

        if os.name == "nt":
            kill_vworks()
            if init_com:
                pythoncom.CoInitializeEx(pythoncom.COINIT_APARTMENTTHREADED)
                logging.info("COM initialized in the driver")
            try:
                self.driver = cc.CreateObject("VWorks4.VWorks4API")
                self.event_sink = VWorksEventSink(self.event_queue)
                self.event_connection = GetEvents(self.driver, self.event_sink)
                self.show_vworks(True)
            except Exception as e:
                logging.error(f"Failed to create VWorks COM object: {e}")
                # Check if this is a COM initialization error
                if "CoInitialize has not been called" in str(e):
                    logging.error(
                        "COM not initialized. Make sure pythoncom.CoInitialize is called first"
                    )
                raise

    def __del__(self) -> None:
        try:
            if hasattr(self, "driver") and self.driver:
                logging.info("Cleaning up BravoServer resources")
                self.close()

            # Uninitialize COM in the same thread that initialized it
            if os.name == "nt":
                import pythoncom

                logging.debug("Uninitializing COM in server thread")
                pythoncom.CoUninitialize()
        except Exception as e:
            logging.error(f"Error during BravoServer cleanup: {e}")

    def login(self, user: str = "administrator", psw: str = "administrator") -> None:
        self.driver.Login(user, psw)
        time.sleep(2)

    def logout(self) -> None:
        if self.driver:
            try:
                self.driver.Logout()
            except Exception as e:
                logging.warning(f"Error during logout: {e}")

    def show_vworks(self, show: bool = True) -> None:
        logging.info(f"Showing VWorks window= {show}")
        self.driver.ShowVWorks(show)

    def run_protocol(self, protocol: str) -> None:
        if os.path.splitext(protocol)[1] != ".pro":
            raise RuntimeError("Invalid file type. Must be .pro")

        if not os.path.exists(protocol):
            raise FileNotFoundError(f"{protocol} does not exist.")

        try:
            pythoncom.CoInitializeEx(pythoncom.COINIT_APARTMENTTHREADED)
            self.show_vworks()
            logging.info(f"Loading protocol: {protocol}")
            self.driver.LoadProtocol(protocol)

            logging.info(f"Running protocol: {protocol}")
            self.driver.RunProtocol(protocol, 1)

            logging.info(f"Waiting for protocol completion: {protocol}")
            success = self.wait_for_protocol_completion(protocol)

            if not success:
                raise RuntimeError(f"Protocol did not complete successfully: {protocol}")

        except Exception as e:
            raise RuntimeError(f"Error running protocol {protocol}: {e}")

    def run_runset(self, runset_file: str) -> None:
        if os.path.splitext(runset_file)[1] != ".rst":
            raise RuntimeError("Invalid file type. Must be .rst")
        if not os.path.exists(runset_file):
            raise FileNotFoundError(f"{runset_file} does not exist.")
        try:
            pythoncom.CoInitializeEx(pythoncom.COINIT_APARTMENTTHREADED)
            self.show_vworks()
            self.driver.LoadRunsetFile(runset_file)
            success = self.wait_for_protocol_completion(runset_file)
            if not success:
                raise RuntimeError(f"Runset did not complete successfully: {runset_file}")
        except Exception as e:
            raise RuntimeError(f"Error running runset {runset_file}:{e}")

    def wait_for_protocol_completion(self, protocol_name: str, timeout: int = 300) -> bool:
        start_time = time.time()
        last_event_type = None
        last_event_data = None
        # Message pump for COM events - keep processing Windows messages
        while True:
            pythoncom.PumpWaitingMessages()
            # Check if we've received completion events
            try:
                # Non-blocking queue check
                event_type, event_data = self.event_queue.get_nowait()
                print(f"Event received: {event_type} - Data: {event_data}")

                last_event_type = event_type
                last_event_data = event_data

                # For matching protocol names
                if event_type == "ProtocolComplete":
                    print(f"Protocol completion event received: {event_data}")
                    return True
                elif event_type == "ProtocolAborted":
                    print(f"Protocol aborted: {event_data}")
                    raise RuntimeError(f"Protocol was aborted: {protocol_name}")
                elif event_type == "UnrecoverableError":
                    print(f"Unrecoverable error occurred: {event_data}")
                    raise RuntimeError(f"Unrecoverable error: {event_data}")
                elif event_type == "RecoverableError":
                    print(f"Recoverable error occurred: {event_data}")
                    # Continue execution for recoverable errors, but log it
            except queue.Empty:
                # No events in the queue, continue waiting
                pass

            # Check timeout
            if time.time() - start_time > timeout:
                # If we timed out, report the last event we received
                error_msg = f"Timed out waiting for protocol completion: {protocol_name}"
                if last_event_type:
                    error_msg += f". Last event was {last_event_type}: {last_event_data}"
                else:
                    error_msg += ". No events were received"
                raise TimeoutError(error_msg)

            # Short sleep to avoid hammering the CPU
            time.sleep(0.28)

    def close(self) -> None:
        """Properly clean up resources"""
        if self.driver:
            self.driver.Logout()
            self.event_connection = None
        if os.name == "nt":
            try:
                if hasattr(self, "driver") and self.driver:
                    try:
                        self.driver.Logout()
                    except Exception as e:
                        logging.warning(
                            f"Driver instanced closed before proper clean up. VWorks might have crashed, {e}"
                        )
                kill_vworks()
                time.sleep(0.5)

                if hasattr(self, "event_connection") and self.event_connection:
                    self.event_connection = None
            except Exception as e:
                logging.warning(f"Error during cleanup: {e}")


# if __name__ == "__main__":
#    # kill_vworks()
#     vworks = None
#     # try:
#     vworks = BravoDriver()
#     vworks.login()
#     vworks.run_runset("C:\\VWorks Workspace\\RunSet Files\\move_to_location_3.rst")
#         # Wait for any final messages to process
#     # except Exception as e:
#     #     print(f"Error running protocol: {e}")
#     # finally:
#     #     if vworks:
#     #         print("Cleaning up resources")
#     #         vworks.close()
#     #     sys.exit(0)



from tools.toolbox.db import Db
from typing import Any, Dict, Union

db = Db()

def get_all_labware() -> Any:
    """Get all labware from the database."""
    response = db.get_data("labware")
    return response

def get_labware(labware_id: Union[int, str]) -> Any:
    """Get a specific labware by ID or name."""
    response = db.get_by_id_or_name(labware_id, "labware")
    return response

def add_labware(labware_data: Dict[str, Any]) -> Any:
    """Add a new labware to the database."""
    response = db.post_data(labware_data, "labware")
    return response

def edit_labware(labware_id: int, labware_data: Dict[str, Any]) -> Any:
    """Edit an existing labware in the database."""
    response = db.update_data(labware_id, labware_data, "labware")
    return response

def delete_labware(labware_id: int) -> Any:
    """Delete a labware from the database."""
    response = db.delete_data(labware_id, "labware")
    return response


import logging
import serial
from typing import Union, Optional
from tools.toolbox.variables import update_variable 
from tools.base_server import ABCToolDriver

READ_TIMEOUT = 5  # Increased timeout for scan operations

def try_utf_decode(data: Union[str, bytes]) -> str:
    """Convert bytes to UTF-8 string if needed."""
    if isinstance(data, str):
        return data
    try:
        return data.decode("utf-8")
    except Exception:
        raise serial.SerialException(f"Error decoding to UTF-8: {data!r}")


class Dataman70Driver(ABCToolDriver):
    def __init__(self, com_port: str, timeout: int = READ_TIMEOUT) -> None:
        """Initialize the barcode scanner driver.
        
        Args:
            com_port: Serial port path (e.g., '/dev/ttyUSB0' or 'COM3')
            timeout: Read timeout in seconds
        """
        self.serial_port = serial.Serial(
            com_port, 
            baudrate=9600,  # Common baudrate for barcode scanners
            timeout=timeout,
            write_timeout=1
        )
        logging.info(f"Connected to Dataman70 on {com_port}")

    def close(self) -> None:
        """Close the serial connection."""
        if self.serial_port.is_open:
            # Disable scanning before closing
            try:
                self.serial_port.write(b"||>SET TRIGGER.TYPE 0\r\n")
                logging.info("Disabled scanning mode")
            except Exception as e:
                logging.error(f"Error disabling scanning mode: {e}")
            
            self.serial_port.close()
            logging.info("Dataman70 connection closed")

    def _read_response(self) -> str:
        """Read a response from the scanner."""
        self.serial_port.reset_input_buffer()
        reply = self.serial_port.read_until(expected=b"\n")
        reply_string = try_utf_decode(reply)
        
        if not reply_string:
            return ""
            
        # Check for proper line ending (should end with \r\n)
        if reply_string.endswith("\r\n"):
            return reply_string[:-2]  # Remove \r\n
        elif reply_string.endswith("\n"):
            return reply_string[:-1]  # Remove \n
        
        return reply_string.strip()


    def assert_barcode(self, expected: str) -> None:
        """Scan and assert that the scanned barcode matches the expected value.
        Args:
            expected: The expected barcode string  
        """
        scanned = self.scan_barcode()
        if scanned is None:
            logging.error("No barcode scanned")
            raise RuntimeError("No barcode detected")
        if scanned == expected:
            logging.info(f"Scanned barcode matches expected: {scanned}")
            return None
        else:
            logging.warning(f"Scanned barcode '{scanned}' does not match expected '{expected}'")
            raise RuntimeError(f"Barcode mismatch: expected '{expected}', got '{scanned}'")

    def scan_barcode(self, mapped_variable: Optional[str]=None) -> Optional[str]:
        """Trigger a barcode scan and return the result.
        
        Returns:
            The scanned barcode as a string, or None if no barcode was scanned
            
        Raises:
            serial.SerialException: If there's a communication error
            Exception: If the scanner is not responding
        """
        try:
            # Enable scanning (trigger type 1)
            self.serial_port.write(b"||>SET TRIGGER.TYPE 1\r\n")
            
            # Read the barcode data
            barcode = self._read_response()
            # Always disable scanning after attempt (trigger type 0)
            self.serial_port.write(b"||>SET TRIGGER.TYPE 0\r\n")
            
            if barcode and len(barcode) > 1:  # Valid barcode (more than 1 character)
                logging.info(f"Scanned barcode: {barcode}")
                if mapped_variable:
                    update_variable(mapped_variable, barcode)
                    logging.info(f"Updated variable '{mapped_variable}' with scanned barcode")
                return barcode
            
            logging.warning("No barcode detected or scan timeout")
            return None
            
        except serial.SerialException as e:
            logging.error(f"Serial communication error: {e}")
            # Try to disable scanning even on error
            try:
                self.serial_port.write(b"||>SET TRIGGER.TYPE 0\r\n")
            except Exception as e:
                raise RuntimeError("Failed to disable scanning after communication error")
            raise
        except Exception as e:
            logging.error(f"Unexpected error during scan: {e}")
            # Try to disable scanning even on error
            try:
                self.serial_port.write(b"||>SET TRIGGER.TYPE 0\r\n")
            except Exception as e:
                raise RuntimeError("Failed to disable scanning after error")
            raise

    def __exit__(self) -> None:
        """Context manager exit - ensures connection is closed."""
        self.close()

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    driver = Dataman70Driver("/dev/tty.usbmodem1A1727PP2448261")  # Update with actual port
    try:
        barcode = driver.scan_barcode("barcsode")
    finally:
        driver.close()



import socket
import time
from typing import Optional

class TcpIp:
    def __init__(self, ip: str, port: int):
        self.ip = ip
        self.port = port
        self.socket: Optional[socket.socket] = None
        self.is_connected = False

    def connect(self) -> None:
        try:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.socket.connect((self.ip, self.port))
            self.is_connected = True
        except Exception as e:
            raise Exception(f"Error establishing connection: {e}")

    def disconnect(self) -> None:
        try:
            if self.socket:
                self.socket.shutdown(socket.SHUT_RDWR)
                self.socket.close()
        except Exception as e:
            raise Exception(f"Error closing connection: {e}")

    def clear_buffer(self) -> None:
        if self.socket:
            self.socket.setblocking(False)
            try:
                while self.socket.recv(1024):
                    pass
            except BlockingIOError:
                pass
            self.socket.setblocking(True)

    def send_command(self, message: str) -> str:
        if self.socket:
            self.socket.sendall(message.encode())
            response = self.socket.recv(1024).decode()
            time.sleep(0.5)
            return response
        return ""

    def read_response(self, buffer_size: int = 1024, timeout: float = 60) -> str:
        if self.socket:
            response_data = bytearray()
            self.socket.settimeout(timeout)
            try:
                while True:
                    chunk = self.socket.recv(buffer_size)
                    response_data.extend(chunk)
                    if len(chunk) < buffer_size:
                        break
                time.sleep(0.5)
                return response_data.decode()
            finally:
                self.socket.settimeout(None)
        return ""


from google.protobuf.struct_pb2 import Struct
import typing as t
import subprocess
import platform

def struct_to_dict(struct: Struct) -> t.Any:
    out = {}
    for key, value in struct.items():
        if isinstance(value, Struct):
            out[key] = struct_to_dict(value)
        else:
            out[key] = value
    return out


def text_to_speech(text: str) -> bool:
    """
    Convert text to speech using native system commands.
    
    Args:
        text (str): The text to speak
    
    Returns:
        bool: True if successful, False otherwise
    """
    
    system = platform.system().lower()
    
    try:
        if system == "darwin":  # macOS
            # Use the 'say' command
            cmd = ["say", text]
            subprocess.run(cmd, check=True)
            return True
            
        elif system == "windows":
            # Use PowerShell with SAPI
            ps_script = (
                f"$s=New-Object -ComObject Sapi.SpVoice;"
                f"$s.Speak(\"{text}\")"
            )
            
            cmd = ["powershell", "-Command", ps_script]
            subprocess.run(cmd, check=True)
            return True
            
        else:
            print(f"Unsupported operating system: {system}")
            return False
            
    except subprocess.CalledProcessError as e:
        print(f"Error running TTS command: {e}")
        return False
    except Exception as e:
        print(f"Unexpected error: {e}")
        return False



import os
from os.path import  dirname
from pydantic import BaseModel
from typing import Optional, Any
from datetime import date , time 
import logging 
import typing as t
from tools.toolbox.workcell import get_all_workcells
from tools.toolbox.db import Db

ROOT_DIRECTORY = dirname(dirname(os.path.realpath(__file__)))

db = Db()

class Tool(BaseModel):
    id: int
    name :str 
    type: str 
    port: int

class WorkcellConfig(BaseModel):
    id:int = 0
    name: str = "workcell_1"
    created_at: Optional[str] = None
    updated_at: Optional[str] = None
    description: Optional[str] = None
    location: Optional[str] = None
    tools: list[Tool] = []

class AppConfig(BaseModel):
    workcell:str
    data_folder:Optional[str]
    host_ip: Optional[str] 
    redis_ip: Optional[str] 
    enable_slack_errors: bool 
    slack_bot_tocken: Optional[str]
    slack_workcell_channel: Optional[str]
    slack_error_channel: Optional[str]
    slack_admins_ids: Optional[list[str]]


def get_workcell(id:int) -> Any:
    response = db.get_by_id_or_name(id, "workcells")
    return response

def get_selected_workcell() -> Any:
    workcell = db.get_data("settings/workcell").get("value")
    return workcell

class Config():
    def __init__(self) -> None:
        self.workcell_config : Optional[WorkcellConfig] = None
        self.workcell_config_file  : str = ""

    def serialize(self, obj:t.Any) -> t.Any:
        """JSON serializer for objects not serializable by default json code"""

        if isinstance(obj, date):
            serial = obj.isoformat()
            return serial

        if isinstance(obj, time):
            serial = obj.isoformat()
            return serial
        return obj.__dict__
    
    def load_workcell_config(self)-> None:
        #Ping the database to check if connection is established
        workcells = None
        selected_workcell = None
        db_is_up = db.ping(1)
        if not db_is_up:
            logging.error("Can't establish connection to galago api.")
            logging.warning("Galago api container might be down. "
                            "No instrument tools will be launched.")
            self.workcell_config = WorkcellConfig()
            return None
        else:
            selected_workcell = get_selected_workcell()
            workcells = get_all_workcells()
            if workcells is None or selected_workcell is None:
                logging.error("No workcells or tools found in the database")
                self.workcell_config = WorkcellConfig()
                return None
            selected_workcell_config = [workcell for workcell 
                                        in workcells if 
                                        workcell.get("name") == selected_workcell][0]
            if selected_workcell:
                self.workcell_config = WorkcellConfig.parse_obj(selected_workcell_config)
        return None
    


import subprocess
import os
import logging
import typing as t
import tempfile
import shutil


def write_to_file(file_name: str, content: str) -> None:
    with open(file_name, "w", encoding="utf-8") as f:
        f.write(content)


def run_python_script(
    script_content: str, 
    blocking: bool = True,
    python_exe: t.Optional[str] = None
) -> t.Optional[str]:

    python_executable = python_exe if python_exe else "python"
    logging.info(f"Using python executable: {python_executable}")
    # Validate the executable exists if a full path was provided
    if python_exe and not os.path.isfile(python_exe):
        raise RuntimeError(f"Python executable not found: {python_exe}")
    
    # Create temporary directory and files
    temp_dir = tempfile.mkdtemp()
    temp_script = os.path.join(temp_dir, "temp_script.py")
    stdout_file = os.path.join(temp_dir, "stdout.txt")
    
    try:
        write_to_file(temp_script, script_content)
        cmd = [python_executable, temp_script]
        
        # Execute script
        with open(stdout_file, 'w', encoding='utf-8') as f:
            process = subprocess.Popen(
                cmd,
                stdout=f,
                stderr=subprocess.STDOUT
            )
            
            if blocking:
                process.wait()
                with open(stdout_file, 'r', encoding='utf-8') as output_f:
                    output = output_f.read()
                if process.returncode != 0:
                    raise RuntimeError(
                        f"Script failed with return code {process.returncode}. Output:\n{output}"
                    )
                
                return output
            
            return None
            
    except FileNotFoundError:
        raise RuntimeError(f"Python executable not found: {python_executable}")
    except Exception as e:
        logging.error(f"Error while running script: {e}")
        raise
    finally:
        # Cleanup temporary files
        logging.info("Cleaning up temporary files")
        try:
            shutil.rmtree(temp_dir)
        except Exception as e:
            logging.warning(f"Failed to cleanup temporary directory: {e}")





import os
import logging 
import threading 
from tools.base_server import ABCToolDriver

try:
    import pythoncom
except Exception:
    # The driver will error if there's no pythoncom
    pass

if os.name == "nt":
    import clr  # type: ignore
    SDK_DLL = os.path.join(
                os.path.dirname(os.path.abspath(__file__)), "deps",
                "AxInterop.MicroplateLabelerLib.dll",
            )

    clr.AddReference(SDK_DLL) 

    from AxMicroplateLabelerLib import AxMicroplateLabeler # type: ignore

else:
    class AxMicroplateLabeler(): # type: ignore
        def __init__(self) -> None:
            pass
    
        def Initialize(self) -> None:
            raise NotImplementedError(
                "AxMicroplateLabeler is not supported on non-Windows platforms"
            )
        
        def Close(self) -> None:
            raise NotImplementedError(
                "AxMicroplateLabeler is not supported on non-Windows platforms"
            )
        
        def HomeStage(self) -> None:
            raise NotImplementedError(
                "AxMicroplateLabeler is not supported on non-Windows platforms"
            )
        
        def PrintAndApply(self) -> None:
            raise NotImplementedError(
                "AxMicroplateLabeler is not supported on non-Windows platforms"
            )
        def PrintAndApplyByFormatName(self) -> None:
            raise NotImplementedError(
                "AxPlateLoc is not supported on non-Windows platforms"
            )
        def PrintLabelByFormatName(self) -> None:
            raise NotImplementedError(
                "AxMicroplateLabeler is not supported on non-Windows platforms"
            )       
        def PrintLabel(self) -> None:
            raise NotImplementedError(
                "AxMicroplateLabeler is not supported on non-Windows platforms"
            )
        def ShowDiagsDialog(self) -> None:
            raise NotImplementedError(
                "AxMicroplateLabeler is not supported on non-Windows platforms"
            )
        def DropStage(self) -> None:
            raise NotImplementedError(
                "AxMicroplateLabeler is not supported on non-Windows platforms"
            )       
        def RotateStage(self) -> None:
            raise NotImplementedError(
                "AxMicroplateLabeler is not supported on non-Windows platforms"
            )
        def Rotate180(self) -> None:
            raise NotImplementedError(
                "AxMicroplateLabeler is not supported on non-Windows platforms"
            )
        def GetLastError(self) -> None:
            raise NotImplementedError(
                "AxMicroplateLabeler is not supported on non-Windows platforms"
            )
SIDE_MAP = {
    'east':1,
    'north':2,
    'west':4,
    'south':8
}

class VCodeDriver(ABCToolDriver):
    def __init__(self , profile:str) -> None:
        self.profile: str = profile
        self.live : bool  = False
        self.client : AxMicroplateLabeler
        self.instantiate()

    def instantiate(self) -> None:
        pythoncom.CoInitialize()
        self.client = AxMicroplateLabeler()
        self.client.CreateControl()
        self.client.Blocking = True
        pythoncom.CoUninitialize()

    def initialize(self) -> None:
        args : dict = {"profile":self.profile}
        self.schedule_threaded_command("initialize", args)

    def close(self) -> None:
        self.schedule_threaded_command("close", {})
    
    def home_stage(self) -> None:
        self.schedule_threaded_command("home_stage",{})

    def print_and_apply_by_index(self, format_index:int, side:str, drop_stage:bool, field_0:str, field_1:str, field_2:str, field_3:str, field_4:str, field_5:str) -> None:
        args :dict = {
            "format_index": format_index,
            "side":int(SIDE_MAP[side]), 
            "drop_stage":drop_stage, 
            "field_0":field_0, 
            "field_1":field_1,
            "field_2":field_2,
            "field_3":field_3,
            "field_4":field_4,
            "field_5":field_5
        }
        self.schedule_threaded_command("print_and_apply", args)

    def print_and_apply_by_name(self, format_name:str, side:str, drop_stage:bool, field_0:str, field_1:str, field_2:str, field_3:str, field_4:str, field_5:str) -> None:
        args :dict = {
            "format_name": format_name,
            "side":int(SIDE_MAP[side]), 
            "drop_stage":bool(drop_stage), 
            "field_0":field_0, 
            "field_1":field_1,
            "field_2":field_2,
            "field_3":field_3,
            "field_4":field_4,
            "field_5":field_5
        }
        self.schedule_threaded_command("print_and_apply_by_name", args)

    def print_label(self, format_index:str, field_0:str, field_1:str, field_2:str, field_3:str, field_4:str, field_5:str) -> None:
        args :dict = {
            "format_index": format_index,
            "field_0":field_0, 
            "field_1":field_1,
            "field_2":field_2,
            "field_3":field_3,
            "field_4":field_4,
            "field_5":field_5
        }
        self.schedule_threaded_command("print_label", args)
    
    def show_diagnostics(self) -> None:
        self.schedule_threaded_command("show_diagnostics", {"modal":True, "level":1})
    
    def drop_stage(self, variant:bool) -> None:
         self.schedule_threaded_command("drop_stage", {"variant":variant})
    
    def rotate_stage(self, angle:float) -> None:
        self.schedule_threaded_command("rotate_stage",{"angle":angle})
    
    def rotate180(self) -> None:
        self.schedule_threaded_command("rotate_180",{})

    def schedule_threaded_command(self, command:str, arguments:dict) -> None:  # type: ignore
        self.execution_thread = threading.Thread(target=self.execute_command(command, arguments,)) # type: ignore
        self.execution_thread.daemon = True
        self.execution_thread.start()
        return None
    
    def execute_command(self, command:str, arguments:dict) -> None:
        response = 0
        try:
            if command == "initialize":
                response = self.client.Initialize(arguments["profile"])
            elif command == "close":
                response  = self.client.Close()
            elif command == "home_stage":
                response = self.client.HomeStage()
            elif command == "print_and_apply":
                response= self.client.PrintAndApply(arguments["format_index"],
                                                 arguments["side"],
                                                 arguments["drop_stage"],
                                                 arguments["field_0"],
                                                 arguments["field_1"],
                                                 arguments["field_2"],
                                                 arguments["field_3"],
                                                 arguments["field_4"],
                                                 arguments["field_5"]
                                                 )
            elif command == "print_and_apply_by_name":
                response= self.client.PrintAndApplyByFormatName(arguments["format_name"],
                                                 arguments["side"],
                                                 arguments["drop_stage"],
                                                 arguments["field_0"],
                                                 arguments["field_1"],
                                                 arguments["field_2"],
                                                 arguments["field_3"],
                                                 arguments["field_4"],
                                                 arguments["field_5"]
                                                 )
            elif command == "print_label":
                response= self.client.PrintLabelByFormatName(arguments["format_name"],
                                                 arguments["field_0"],
                                                 arguments["field_1"],
                                                 arguments["field_2"],
                                                 arguments["field_3"],
                                                 arguments["field_4"],
                                                 arguments["field_5"]
                                                 )

            elif command == "show_diagnostics":
                response = self.client.ShowDiagsDialog(arguments["modal"],arguments["level"])
            elif command == "drop_stage":
                response= self.client.DropStage(arguments["variant"])
            elif command == "rotate_stage":
                response= self.client.RotateStage(arguments["angle"])
            elif command == "rotate_180":
                response= self.client.Rotate180()
            elif command == "get_last_error":
                response= self.client.GetLastError()
            else:
                response = -1
        except RuntimeError as e:
            self.live = False
            raise RuntimeError(f"Failed to execute command {str(e)}")
        finally:
            pythoncom.CoUninitialize()
            logging.info(f"Received response is {response}")
            if(response != 0):
                error : str = self.client.GetLastError()
                raise RuntimeError(f"Failed to execute command {command} with error {error}")
            return None


import logging
from tools.base_server import ToolServer, serve
from tools.grpc_interfaces.pyhamilton_pb2  import Command, Config
from tools.grpc_interfaces.tool_base_pb2 import ExecuteCommandReply
import argparse 
from google.protobuf.struct_pb2 import Struct
from tools.grpc_interfaces.tool_base_pb2 import  SUCCESS, ERROR_FROM_TOOL
from tools.toolbox.python_subprocess import run_python_script

class PyHamiltonServer(ToolServer):
    toolType = "pyhamilton"
    config:Config

    def __init__(self) -> None:
        super().__init__()


    def _configure(self, request:Config) -> None:
        logging.info("Configuring PyHamilton...")
        self.config = request
        return
        
        
    def RunScript(self, params:Command.RunScript) -> ExecuteCommandReply:
        s  = Struct()
        response = ExecuteCommandReply()
        response.return_reply = True
        response.response = SUCCESS
        try:
            result = run_python_script(params.script_content, blocking=True)
            logging.info(f"Script result is {result}")
            if response:
                s.update({'response':result})
            else:
                s.update({'response':''})
            response.meta_data.CopyFrom(s)
        except Exception as exc:
            logging.exception(exc)
            response.response = ERROR_FROM_TOOL
            response.error_message = str(exc)
        return response

    def RunLocalScript(self, params:Command.RunLocalScript) -> ExecuteCommandReply:
        s  = Struct()
        response = ExecuteCommandReply()
        response.return_reply = True
        response.response = SUCCESS
        try:
            if not params.path:
                raise ValueError("Path to script must be provided...")
            python_exe = None
            logging.info(f"Config is {self.config}")
            if self.config and self.config.python_exe:
                python_exe = self.config.python_exe
                logging.info(f"Using python executable from config: {python_exe}")
            result = run_python_script(f"exec(open(r'''{params.path}''').read())", blocking=True, python_exe=python_exe)
            logging.info(f"Script result is {result}")
            if response:
                s.update({'response':result})
            else:
                s.update({'response':''})
            response.meta_data.CopyFrom(s)
        except Exception as exc:
            logging.exception(exc)
            response.response = ERROR_FROM_TOOL
            response.error_message = str(exc)
        return response
     
             
if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument('--port')
    args = parser.parse_args()
    if not args.port:
         raise RuntimeWarning("Port must be provided...")
    serve(PyHamiltonServer(), str(args.port))



from tools.toolbox.db import Db
from typing import Any, Dict, Union

db = Db()

def get_all_protocols() -> Any:
    """Get all protocols from the database."""
    response = db.get_data("protocols")
    return response

def get_protocols_by_workcell(workcell_name: str) -> Any:
    """Get all protocols for a specific workcell."""
    response = db.get_data(f"protocols?workcell_name={workcell_name}")
    return response

def get_protocol(protocol_id: Union[int, str]) -> Any:
    """Get a specific protocol by ID."""
    response = db.get_by_id_or_name(protocol_id, "protocols")
    return response

def create_protocol(protocol_data: Dict[str, Any]) -> Any:
    """Create a new protocol in the database.
    
    Args:
        protocol_data: Dictionary containing protocol information including:
            - name: Protocol name
            - category: Protocol category
            - workcell_id: ID of the associated workcell
            - description: Optional description
            - icon: Optional icon
            - params: Parameters for the protocol
            - commands: List of commands for the protocol
            - version: Optional version number (defaults to 1)
            - is_active: Optional active status (defaults to True)
    """
    # Set default values if not provided
    if "version" not in protocol_data:
        protocol_data["version"] = 1
    if "is_active" not in protocol_data:
        protocol_data["is_active"] = True
    if "params" not in protocol_data:
        protocol_data["params"] = {}
    if "commands" not in protocol_data:
        protocol_data["commands"] = []
        
    response = db.post_data(protocol_data, "protocols")
    return response

def update_protocol(protocol_id: int, protocol_data: Dict[str, Any]) -> Any:
    """Update an existing protocol in the database."""
    response = db.update_data(protocol_id, protocol_data, "protocols")
    return response

def delete_protocol(protocol_id: int) -> Any:
    """Delete a protocol from the database."""
    response = db.delete_data(protocol_id, "protocols")
    return response



import os
import sys
import argparse
import logging
import subprocess
from tools import __version__ as galago_version

# Move heavy imports inside functions to delay loading
def get_shell_command(tool: str, file: str) -> list[str]:
    if tool:
        return [sys.executable, '-m', f'tools.{tool}.server']
    elif file:
        return [sys.executable, file]
    else:
        raise RuntimeError("Either tool or file must be provided.")

def serve() -> None:
    """Legacy serve function for backwards compatibility"""
    parser = argparse.ArgumentParser()
    parser.add_argument('--port', required=True, help="Port must be provided.")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--tool', help="Name of the tool to run (e.g., 'xpeel').")
    group.add_argument('--file', help="Path to the file to run.")
    known, remaining = parser.parse_known_args()
    
    use_shell = os.name == 'nt'
    sys.argv = [sys.argv[0]] + remaining
    command = get_shell_command(str(known.tool).lower(), known.file)
    command.append(f'--port={known.port}')
    
    # Set the PYTHONPATH to include the project root.
    env = os.environ.copy()
    project_root = os.getcwd()
    current_pythonpath = env.get("PYTHONPATH", "")
    env["PYTHONPATH"] = project_root + (os.pathsep + current_pythonpath if current_pythonpath else "")
    
    subprocess.Popen(command, shell=use_shell, universal_newlines=True, env=env)

def start_web_server() -> bool:
    """Start the web socket server in a separate process"""
    try:
        # Import here to avoid import issues
        from tools.web_server import main as web_server_main
        import asyncio
        
        # Run the web server in the current process
        asyncio.run(web_server_main())
    except Exception as e:
        print(f"Failed to start web server: {e}")
        return False
    return True

def launch_web_only() -> int:
    """Launch only the web server without desktop app"""
    try:
        import asyncio
        from tools.web_server import main as web_server_main
        asyncio.run(web_server_main())
        return 0
    except Exception as e:
        print(f"Failed to start web server: {e}")
        return 1

def main() -> None:
    """Main entry point for Galago Tools Manager CLI"""
    
    # Define top-level arguments
    parser = argparse.ArgumentParser(
        description="Galago Tools Manager - Modern Lab Automation Interface",
        add_help=False
    )
    parser.add_argument("--help", action="help", help="Show this help message and exit")
    parser.add_argument("--version", action="version", version=f"Galago Tools Manager {galago_version}", help="Show program version and exit")
    parser.add_argument("--discover", action="store_true", help="Autodiscover tools")
    parser.add_argument("--console", action="store_true", help="Launch in headless mode. No GUI.")
    parser.add_argument("--web-only", action="store_true", help="Launch web server only (no desktop app)")
    parser.add_argument("--debug", action="store_true", help="Enable debug logging")
    parser.add_argument("--legacy", action="store_true", help="Launch legacy app.")
    parser.add_argument("--list", action="store_true", help="List available tools")
    parser.add_argument("--info", metavar="TOOL", help="Get information about a specific tool")
    
    # Parse known arguments and get the remaining arguments (if any)
    known, remaining = parser.parse_known_args()
    
    # Set up logging
    level = logging.DEBUG if known.debug else logging.INFO
    logging.basicConfig(
        level=level,
        format='%(asctime)s | %(levelname)s | %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )
    
    # Remove the consumed arguments from sys.argv
    sys.argv = [sys.argv[0]] + remaining
    
    # Handle legacy modes
    if known.console:
        from tools.launch_console import main as launch_console_main
        sys.exit(launch_console_main())
    elif known.legacy:
        from tools.launch_tools import main as launch_tools_main
        sys.exit(launch_tools_main())
    elif known.discover:
        from tools.discover_tools import main as autodiscover_main
        sys.exit(autodiscover_main())
    elif known.web_only:
        # Launch web server only
        sys.exit(launch_web_only())
    elif known.list:
        from tools.utils import list_available_tools
        tools = list_available_tools()
        print("Available tools:")
        for tool in tools:
            print(f"- {tool}")
        sys.exit(0)
    elif known.info:
        from tools.utils import print_tool_server_info
        print_tool_server_info(str(known.info).lower())
        sys.exit(0)
    else:
        # Default: Lanch tools manager web server
        try:
            from tools.web_server import main as web_server_main
            import asyncio
            asyncio.run(web_server_main())
        except KeyboardInterrupt:
            print("Shutting down...")
            sys.exit(0)
        except Exception as e:
            logging.error(f"Failed to start Galago Tools Manager: {e}")
            sys.exit(1)

if __name__ == "__main__":
    main()


import socket
import grpc
import concurrent.futures
import time
import os
import datetime
from grpc_reflection.v1alpha import reflection_pb2, reflection_pb2_grpc

def is_port_open(host: str, port: int, timeout: float = 1.0) -> bool:
    """Quick check if a TCP port is open on the host."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.settimeout(timeout)
        try:
            sock.connect((host, port))
            return True
        except Exception:
            return False

def check_grpc_reflection(host: str, port: int) -> bool:
    """
    Attempt to call the gRPC reflection API.
    
    Returns True if the server responds to a reflection request.
    Note: The server must have reflection enabled.
    """
    target = f"{host}:{port}"
    channel = grpc.insecure_channel(target)
    stub = reflection_pb2_grpc.ServerReflectionStub(channel)
    request = reflection_pb2.ServerReflectionRequest(list_services="")

    try:
        responses = stub.ServerReflectionInfo(iter([request]))
        for response in responses:
            if response.HasField("list_services_response"):
                return True
    except Exception:
        return False
    finally:
        channel.close()  # Ensure the channel is closed after use.
    return False

def discover_grpc_servers(host: str, start_port: int, end_port: int) -> list[int]:
    """
    Scan a range of ports on the given host.
    
    Returns a list of ports where a gRPC server (with reflection enabled)
    was detected.
    """
    found_ports = []
    with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
        future_to_port = {}
        for port in range(start_port, end_port + 1):
            if is_port_open(host, port):
                # Submit the reflection check for this open port.
                future = executor.submit(check_grpc_reflection, host, port)
                future_to_port[future] = port

        for future in concurrent.futures.as_completed(future_to_port):
            port = future_to_port[future]
            try:
                if future.result():
                    found_ports.append(port)
            except Exception:
                pass
    return found_ports

def main() -> int:
    host = "localhost"
    # the scanning range is defined on db service.
    start_port = 4000
    end_port = 5000
    hardcoded_port = 1010  # for tool box

    try:
        while True:
            os.system('cls' if os.name == 'nt' else 'clear')
            timestamp = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
            print(f"==== Galago Tools gRPC Server Discovery at {timestamp} ====\n")
            print(f"Scanning {host} ports {start_port} to {end_port} for gRPC servers...\n")
            
            discovered_ports = set(discover_grpc_servers(host, start_port, end_port))
            if is_port_open(host, hardcoded_port) and check_grpc_reflection(host, hardcoded_port):
                discovered_ports.add(hardcoded_port)
            
            if discovered_ports:
                print("Discovered gRPC servers on the following ports:")
                for port in sorted(discovered_ports):
                    print(f"  - {host}:{port}")
            else:
                print("No active gRPC servers found.")

            print("\nNext scan in 10 seconds. Press Ctrl+C to exit.")
            time.sleep(10)
    except KeyboardInterrupt:
        print("\nDiscovery interrupted by user. Exiting gracefully.")
        return 0 
    except Exception as e:
        print(f"An error occurred: {e}")
        return 1

if __name__ == "__main__":
    main()
